import traceback
from typing import Dict, List, Optional

from supervisely._utils import logger
from supervisely.app import StateJson
from supervisely.app.content import DataJson
from supervisely.app.widgets import Widget


class RadioTabs(Widget):
    """Tabs-like widget that switches between panes and notifies on selection changes."""

    class Routes:
        """Callback route names used by the widget frontend to notify Python."""

        VALUE_CHANGED = "value_changed_cb"

    class RadioTabPane:
        """One selectable tab pane (title + content widget + optional subtitle)."""

        def __init__(
            self,
            title: str,
            content: Widget,
            subtitle: Optional[str] = "",
        ):
            """:param title: Tab title.
            :type title: str
            :param content: Widget to display when tab is active.
            :type content: Widget
            :param subtitle: Optional subtitle.
            :type subtitle: str, optional
            """
            self.title = title
            self.subtitle = subtitle
            self.name = title  # identifier corresponding to the active tab
            self.content = content

    def __init__(
        self,
        titles: List[str],
        contents: List[Widget],
        descriptions: Optional[List[str]] = None,
        widget_id=None,
    ):
        """:param titles: List of tab titles (unique, max 10).
        :type titles: List[str]
        :param contents: List of widgets, one per tab.
        :type contents: List[Widget]
        :param descriptions: Optional descriptions per tab.
        :type descriptions: List[str], optional
        :param widget_id: Unique widget identifier.

        :raises ValueError: If titles/contents lengths differ, or titles not unique.
        """
        if len(titles) != len(contents):
            raise ValueError(
                "titles length must be equal to contents length in RadioTabs widget."
            )
        if len(titles) > 10:
            raise ValueError("You can specify up to 10 tabs.")
        if descriptions is None:
            descriptions = [""] * len(titles)
        else:
            if len(titles) != len(descriptions):
                raise ValueError(
                    "descriptions length must be equal to titles length in RadioTabs widget."
                )
        if len(set(titles)) != len(titles):
            raise ValueError("All of tab labels should be unique.")
        self._items = []
        for title, widget, description in zip(titles, contents, descriptions):
            self._items.append(
                RadioTabs.RadioTabPane(
                    title=title, content=widget, subtitle=description
                )
            )
        self._value = titles[0]
        self._changes_handled = False
        super().__init__(widget_id=widget_id, file_path=__file__)

    def value_changed(self, func):
        route_path = self.get_route_path(RadioTabs.Routes.VALUE_CHANGED)
        server = self._sly_app.get_server()
        self._changes_handled = True

        @server.post(route_path)
        def _value_changed():
            res = self.get_active_tab()
            func(res)

        return _value_changed

    def get_json_data(self) -> Dict:
        return {"tabsOptions": {item.name: {"disabled": False} for item in self._items}}

    def get_json_state(self) -> Dict:
        return {"value": self._value}

    def set_active_tab(self, value: str):
        self._value = value
        StateJson()[self.widget_id]["value"] = self._value
        StateJson().send_changes()

    def get_active_tab(self) -> str:
        return StateJson()[self.widget_id]["value"]

    def disable_tab(self, tab_name: str):
        if tab_name not in [item.name for item in self._items]:
            raise ValueError(f"Tab with name '{tab_name}' does not exist.")
        DataJson()[self.widget_id]["tabsOptions"][tab_name]["disabled"] = True
        DataJson().send_changes()

    def enable_tab(self, tab_name: str):
        if tab_name not in [item.name for item in self._items]:
            raise ValueError(f"Tab with name '{tab_name}' does not exist.")
        DataJson()[self.widget_id]["tabsOptions"][tab_name]["disabled"] = False
        DataJson().send_changes()



import random
import time
from collections.abc import Sized


class DynamicSampler:
    """
    A sampler that dynamically adjusts to the size of a dataset that grows over time.
    Implements torch.utils.data.Sampler interface.
    """

    def __init__(self, dataset: Sized, shuffle: bool = True, seed: int = 0):
        """
        :param dataset: Sized dataset.
        :type dataset: Sized
        :param shuffle: Shuffle indices.
        :type shuffle: bool
        :param seed: Random seed.
        :type seed: int
        """
        self.dataset = dataset
        self.shuffle = shuffle
        self.seed = seed

    def __iter__(self):
        remaining_indices = []
        last_known_len = 0

        # Wait for first samples
        while len(self.dataset) == 0:
            time.sleep(0.1)

        while True:
            current_len = len(self.dataset)

            if current_len > last_known_len:
                new_indices = list(range(last_known_len, current_len))
                if self.shuffle:
                    random.shuffle(new_indices)
                remaining_indices.extend(new_indices)
                last_known_len = current_len

            if not remaining_indices:
                # Reshuffle existing data
                remaining_indices = list(range(current_len))
                if self.shuffle:
                    random.shuffle(remaining_indices)

            yield remaining_indices.pop()

    def __len__(self):
        return len(self.dataset)



# coding: utf-8
"""Work with volume figures via the Supervisely API."""

import os
import re
import tempfile
from typing import Dict, List, Literal, Optional
from uuid import UUID

from numpy import uint8
from requests_toolbelt import MultipartDecoder, MultipartEncoder

import supervisely.volume_annotation.constants as constants
from supervisely._utils import batched
from supervisely.api.entity_annotation.figure_api import FigureApi, FigureInfo
from supervisely.api.module_api import ApiField
from supervisely.geometry.closed_surface_mesh import ClosedSurfaceMesh
from supervisely.geometry.mask_3d import Mask3D
from supervisely.io.fs import ensure_base_path, file_exists
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.volume.nrrd_encoder import encode
from supervisely.volume_annotation.plane import Plane
from supervisely.volume_annotation.volume_figure import VolumeFigure


class VolumeFigureApi(FigureApi):
    """
    API for working with :class:`~supervisely.volume_annotation.volume_figure.VolumeFigure`.
    :class:`~supervisely.api.volume.volume_figure_api.VolumeFigureApi` object is immutable.
    """

    def create(
        self,
        volume_id: int,
        object_id: int,
        plane_name: str,
        slice_index: int,
        geometry_json: dict,
        geometry_type: str,
        custom_data: Optional[dict] = None,
        # track_id=None,
    ):
        """
        Create new VolumeFigure for given slice in given volume ID.

        :param volume_id: Volume ID in Supervisely.
        :type volume_id: int
        :param object_id: ID of the object to which the VolumeFigure belongs.
        :type object_id: int
        :param plane_name: Plane of the slice in volume.
        :type plane_name: str
        :param slice_index: Number of the slice to add :class:`~supervisely.volume_annotation.volume_figure.VolumeFigure`.
        :type slice_index: int
        :param geometry_json: Parameters of geometry for :class:`~supervisely.volume_annotation.volume_figure.VolumeFigure`.
        :type geometry_json: dict
        :param geometry_type: Type of VolumeFigure geometry.
        :type geometry_type: str
        :returns: New figure ID
        :rtype: int

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly
                from supervisely.volume_annotation.plane import Plane

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                volume_id = 19581134
                object_id = 5565016
                slice_index = 0
                plane_name = Plane.AXIAL
                geometry_json = {'points': {'exterior': [[500, 500], [1555, 1500]], 'interior': []}}
                geometry_type = 'rectangle'

                figure_id = api.volume.figure.create(
                    volume_id,
                    object_id,
                    plane_name,
                    slice_index,
                    geometry_json,
                    geometry_type
                ) # 87821207
        """

        Plane.validate_name(plane_name)

        return super().create(
            volume_id,
            object_id,
            {
                constants.SLICE_INDEX: slice_index,
                constants.NORMAL: Plane.get_normal(plane_name),
                # for backward compatibility
                ApiField.META: {
                    constants.SLICE_INDEX: slice_index,
                    constants.NORMAL: Plane.get_normal(plane_name),
                },
            },
            geometry_json,
            geometry_type,
            custom_data=custom_data,
            # track_id,
        )

    def append_bulk(self, volume_id: int, figures: List[VolumeFigure], key_id_map: KeyIdMap):
        """
        Add VolumeFigures to given Volume by ID.

        :param volume_id: Volume ID in Supervisely.
        :type volume_id: int
        :param key_id_map: KeyIdMap object.
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`
        :param figures: List of VolumeFigure objects.
        :type figures: List[:class:`~supervisely.volume_annotation.volume_figure.VolumeFigure`]
        :returns: None
        :rtype: None

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly
                from supervisely.volume_annotation.plane import Plane

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                project_id = 19370
                volume_id = 19617444

                key_id_map = sly.KeyIdMap()

                project_meta_json = api.project.get_meta(project_id)
                project_meta = sly.ProjectMeta.from_json(project_meta_json)

                vol_ann_json = api.volume.annotation.download(volume_id)
                vol_ann = sly.VolumeAnnotation.from_json(vol_ann_json, project_meta, key_id_map)
                volume_obj_collection = vol_ann.objects.to_json()
                vol_obj = sly.VolumeObject.from_json(volume_obj_collection[1], project_meta)

                figure = sly.VolumeFigure(
                    vol_obj,
                    sly.Rectangle(20, 20, 129, 200),
                    sly.Plane.AXIAL,
                    45,
                )

                api.volume.figure.append_bulk(volume_id, [figure], key_id_map)
        """

        if len(figures) == 0:
            return

        keys = []
        mask3d_keys = []
        figures_json = []
        mask3d_figures = []

        for figure in figures:
            if figure.geometry.name() == Mask3D.name():
                mask3d_keys.append(figure.key())
                mask3d_figures.append(figure)
            else:
                keys.append(figure.key())
                figures_json.append(figure.to_json(key_id_map, save_meta=True))
        # Figure is missing required field \"meta.normal\"","index":0}}
        self._append_bulk(volume_id, figures_json, keys, key_id_map)
        if len(mask3d_figures) != 0:
            self._append_bulk_mask3d(volume_id, mask3d_figures, mask3d_keys, key_id_map)

    def _download_geometries_batch(self, ids: List[int]):
        """
        Private method. Download figures geometries with given IDs from storage.
        """

        for batch_ids in batched(ids):
            response = self._api.post("figures.bulk.download.geometry", {ApiField.IDS: batch_ids})
            decoder = MultipartDecoder.from_response(response)
            for part in decoder.parts:
                content_utf8 = part.headers[b"Content-Disposition"].decode("utf-8")
                # Find name="1245" preceded by a whitespace, semicolon or beginning of line.
                # The regex has 2 capture group: one for the prefix and one for the actual name value.
                figure_id = int(re.findall(r'(^|[\s;])name="(\d*)"', content_utf8)[0][1])
                yield figure_id, part

    def download_stl_meshes(self, ids: List[int], paths: List[str]):
        """
        Download STL meshes for the specified figure IDs and saves them to the specified paths.

        :param ids: VolumeFigure ID in Supervisely.
        :type ids: int
        :param paths: List of paths to download.
        :type paths: List[str]
        :returns: None
        :rtype: None

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                STORAGE_DIR = sly.app.get_data_dir()

                volume_id = 19371414
                project_id = 17215

                volume = api.volume.get_info_by_id(volume_id)

                key_id_map = sly.KeyIdMap()
                project_meta_json = api.project.get_meta(project_id)
                project_meta = sly.ProjectMeta.from_json(project_meta_json)

                vol_ann_json = api.volume.annotation.download(volume_id)
                id_to_paths = {}
                vol_ann = sly.VolumeAnnotation.from_json(vol_ann_json, project_meta, key_id_map)

                for sp_figure in vol_ann.spatial_figures:
                    figure_id = key_id_map.get_figure_id(sp_figure.key())
                    id_to_paths[figure_id] = f"{STORAGE_DIR}/{sp_figure.key().hex}.stl"

                if id_to_paths:
                    api.volume.figure.download_stl_meshes(*zip(*id_to_paths.items()))
        """

        if len(ids) == 0:
            return
        if len(ids) != len(paths):
            raise RuntimeError('Can not match "ids" and "paths" lists, len(ids) != len(paths)')

        id_to_path = {id: path for id, path in zip(ids, paths)}
        for img_id, resp_part in self._download_geometries_batch(ids):
            ensure_base_path(id_to_path[img_id])
            with open(id_to_path[img_id], "wb") as w:
                w.write(resp_part.content)

    def interpolate(self, volume_id: int, spatial_figure: VolumeFigure, key_id_map: KeyIdMap):
        """
        Interpolate a spatial figure with a ClosedSurfaceMesh geometry.

        :param volume_id: VolumeFigure ID in Supervisely.
        :type volume_id: int
        :param spatial_figure: Spatial figure to interpolate.
        :type spatial_figure: :class:`~supervisely.volume_annotation.volume_figure.VolumeFigure`
        :param key_id_map: KeyIdMap object.
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`
        :returns: None
        :rtype: None

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                volume_id = 19371414
                project_id = 17215

                volume = api.volume.get_info_by_id(volume_id)

                key_id_map = sly.KeyIdMap()
                project_meta_json = api.project.get_meta(project_id)
                project_meta = sly.ProjectMeta.from_json(project_meta_json)

                vol_ann_json = api.volume.annotation.download(volume_id)
                id_to_paths = {}
                vol_ann = sly.VolumeAnnotation.from_json(vol_ann_json, project_meta, key_id_map)

                for sp_figure in vol_ann.spatial_figures:
                    res = volume_figure_api.interpolate(volume_id, sp_figure, key_id_map)
        """

        if type(spatial_figure._geometry) != ClosedSurfaceMesh:
            raise TypeError(
                "Interpolation can be created only for figures with geometry ClosedSurfaceMesh"
            )
        object_id = key_id_map.get_object_id(spatial_figure.volume_object.key())
        response = self._api.post(
            "figures.volumetric_interpolation",
            {ApiField.VOLUME_ID: volume_id, ApiField.OBJECT_ID: object_id},
        )
        return response.content

    # def interpolate_batch(
    #     self, volume_id, spatial_figures: List[VolumeFigure], key_id_map: KeyIdMap
    # ):
    #     # raise NotImplementedError()
    #     # STL mesh interpolations can not be uploaded:
    #     # 400 Client Error: Bad Request for url: public/api/v3/figures.bulk.add ({"error":"Please, use \"figures.bulk.upload.geometry\" method to update figures with \"geometryType\" closed_surface_mesh","details":{"figures":[14]}})
    #     # figures.bulk.upload.geometry

    #     meshes = []
    #     for mesh in spatial_figures:
    #         object_id = key_id_map.get_object_id(mesh.volume_object.key())
    #         response = self._api.post(
    #             "figures.volumetric_interpolation",
    #             {ApiField.VOLUME_ID: volume_id, ApiField.OBJECT_ID: object_id},
    #         )
    #         figure_id = key_id_map.get_figure_id(mesh.key())

    #         # @TODO: load from disk or get from server
    #         meshes.append(response.json())

    #     # for batch in batched(items_to_upload):
    #     #     content_dict = {}
    #     #     for idx, item in enumerate(batch):
    #     #         content_dict["{}-file".format(idx)] = (str(idx), func_item_to_byte_stream(item), 'nrrd/*')
    #     #     encoder = MultipartEncoder(fields=content_dict)
    #     #     self._api.post('import-storage.bulk.upload', encoder)
    #     #     if progress_cb is not None:
    #     #         progress_cb(len(batch))

    #     # content_dict = {}
    #     # for idx, item in enumerate(meshes):
    #     #     content_dict[f"{idx}-mesh"] = (
    #     #         str(idx),
    #     #         func_item_to_byte_stream(item),
    #     #     )
    #     # encoder = MultipartEncoder(fields=content_dict)

    #     # self._api.post("figures.bulk.upload.geometry", encoder)

    #     # # if progress_cb is not None:
    #     # #     progress_cb(len(batch))

    #     # #     self._api.post(
    #     # #         "figures.bulk.upload.geometry",
    #     # #         {ApiField.FIGURE_ID: figure_id, ApiField.GEOMETRY: response.json()},
    #     # #     )

    #     # #     results.append(response.json())
    #     # return results

    def _upload_meshes_batch(self, figure2bytes):
        """
        Private method. Upload figures geometry by given ID to storage.

        :param figure2bytes: Dictionary with figures IDs and geometries.
        :type figure2bytes: dict
        :rtype: None
        """

        for figure_id, figure_bytes in figure2bytes.items():
            content_dict = {
                ApiField.FIGURE_ID: str(figure_id),
                ApiField.GEOMETRY: (str(figure_id), figure_bytes, "application/sla"),
            }
            encoder = MultipartEncoder(fields=content_dict)
            resp = self._api.post("figures.bulk.upload.geometry", encoder)

    def upload_stl_meshes(
        self,
        volume_id: int,
        spatial_figures: List[VolumeFigure],
        key_id_map: KeyIdMap,
        interpolation_dir=None,
    ):
        """
        Upload existing interpolations or create on the fly and and add them to empty mesh figures.

        :param volume_id: VolumeFigure ID in Supervisely.
        :type volume_id: int
        :param spatial_figures: List of spatial figures to upload.
        :type spatial_figures: List[:class:`~supervisely.volume_annotation.volume_figure.VolumeFigure`]
        :param key_id_map: KeyIdMap object.
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`
        :returns: None
        :rtype: None

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                volume_id = 19371414
                project_id = 17215

                volume = api.volume.get_info_by_id(volume_id)

                key_id_map = sly.KeyIdMap()
                project_meta_json = api.project.get_meta(project_id)
                project_meta = sly.ProjectMeta.from_json(project_meta_json)

                vol_ann_json = api.volume.annotation.download(volume_id)
                id_to_paths = {}
                vol_ann = sly.VolumeAnnotation.from_json(vol_ann_json, project_meta, key_id_map)
                sp_figures = vol_ann.spatial_figures

                res = volume_figure_api.upload_stl_meshes(volume_id, sp_figures, key_id_map)
        """

        if len(spatial_figures) == 0:
            return

        figure2bytes = {}
        for sp in spatial_figures:
            figure_id = key_id_map.get_figure_id(sp.key())
            if interpolation_dir is not None:
                meth_path = os.path.join(interpolation_dir, sp.key().hex + ".stl")
                if file_exists(meth_path):
                    with open(meth_path, "rb") as in_file:
                        meth_bytes = in_file.read()
                    figure2bytes[figure_id] = meth_bytes
            # else - no stl file
            if figure_id not in figure2bytes:
                meth_bytes = self.interpolate(volume_id, sp, key_id_map)
                figure2bytes[figure_id] = meth_bytes
        self._upload_meshes_batch(figure2bytes)

    def _append_bulk_mask3d(
        self,
        entity_id: int,
        figures: List[VolumeFigure],
        figures_keys: List,
        key_id_map: KeyIdMap,
        field_name=ApiField.ENTITY_ID,
    ):
        """The same method as _append_bulk but for spatial figures. Uploads figures to given Volume by ID.
        You need to upload the geometry right after figures will be created

        :param entity_id: Volume ID.
        :type entity_id: int
        :param figures_json: List of figure dicts.
        :type figures_json: list
        :param figures_keys: List of figure keys as UUID.
        :type figures_keys: list
        :param key_id_map: KeyIdMap object (dict with bidict values)
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`
        :param field_name: field name for request body
        :type field_name: str
        :rtype: None
        """

        if len(figures) == 0:
            return

        empty_figures = []
        for figure in figures:
            empty_figures.append(
                {
                    ApiField.OBJECT_ID: key_id_map.get_object_id(figure.volume_object.key()),
                    ApiField.GEOMETRY_TYPE: Mask3D.name(),
                    ApiField.LABELING_TOOL: Mask3D.name(),
                    ApiField.ENTITY_ID: entity_id,
                    ApiField.CUSTOM_DATA: figure.custom_data
                }
            )
        for batch_keys, batch_jsons in zip(
            batched(figures_keys, batch_size=100), batched(empty_figures, batch_size=100)
        ):
            resp = self._api.post(
                "figures.bulk.add",
                {field_name: entity_id, ApiField.FIGURES: batch_jsons},
            )
            for key, resp_obj in zip(batch_keys, resp.json()):
                figure_id = resp_obj[ApiField.ID]
                key_id_map.add_figure(key, figure_id)

                for figure in figures:
                    if figure.key() == key:
                        geometry_data = figure.geometry.data
                        header = figure.geometry.create_header()
                        geometry_bytes = encode(geometry_data.astype(uint8), header)
                        self.upload_sf_geometries([key], {key: geometry_bytes}, key_id_map)

    def upload_sf_geometries(
        self,
        spatial_figures: List[UUID],
        geometries: Dict[UUID, bytes],
        key_id_map: KeyIdMap,
    ):
        """
        Upload geometries as bytes into spatial figures in the project using their UUID keys.

        :param spatial_figures: List of UUID keys representing spatial figures.
        :type spatial_figures: List[UUID]
        :param geometries: Dictionary where keys are UUIDs of spatial figures, and values are geometries represented as NRRD files in byte format.
        :type geometries: Dict[UUID, bytes]
        :param key_id_map: KeyIdMap object (a dictionary with bidict values).
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`
        :returns: None
        :rtype: None

        :Usage Example:

            .. code-block:: python

                import os
                import numpy as np
                from dotenv import load_dotenv

                import supervisely as sly
                from supervisely.volume.nrrd_encoder import encode

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                volume_id = 23772225
                project_id = 28159
                geometries = {}
                key_id_map = sly.KeyIdMap()
                geometry_bytes = encode(np.random.randint(2, size=(20, 20, 20), dtype=np.uint8))

                vol_ann_json = api.volume.annotation.download(volume_id)
                project_meta = sly.ProjectMeta.from_json(api.project.get_meta(project_id))
                ann = sly.VolumeAnnotation.from_json(vol_ann_json, project_meta, key_id_map)
                spatial_figures = [sp_figure.key() for sp_figure in ann.spatial_figures]
                for figure in spatial_figures:
                    geometries[figure] = geometry_bytes
                api.volume.figure.upload_sf_geometries(spatial_figures, geometries, key_id_map)
        """

        if len(spatial_figures) == 0:
            return

        for sf in spatial_figures:
            figure_id = key_id_map.get_figure_id(sf)
            geometry_bytes = geometries.get(sf)
            content_dict = {
                ApiField.FIGURE_ID: str(figure_id),
                ApiField.GEOMETRY: (str(figure_id), geometry_bytes, "application/sla"),
            }
            encoder = MultipartEncoder(fields=content_dict)
            self._api.post("figures.bulk.upload.geometry", encoder)

    def upload_sf_geometry(self, spatial_figures: Dict, geometries: List, key_id_map: KeyIdMap):
        """
        Upload spatial figures geometry as bytes to storage by given ID.

        :param spatial_figures: Dictionary with figure IDs.
        :type spatial_figures: dict
        :param geometries: Dictionary with geometries, which represented as content of NRRD files in byte format.
        :type geometries: list
        :param key_id_map: KeyIdMap object (dict with bidict values)
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`
        :rtype: None
        """

        for sf, geometry_bytes in zip(spatial_figures, geometries):
            figure_id = key_id_map.get_figure_id(sf.key())
            content_dict = {
                ApiField.FIGURE_ID: str(figure_id),
                ApiField.GEOMETRY: (str(figure_id), geometry_bytes, "application/sla"),
            }
            encoder = MultipartEncoder(fields=content_dict)
            self._api.post("figures.bulk.upload.geometry", encoder)

    def download_sf_geometries(self, ids: List[int], paths: List[str]):
        """
        Download spatial figure geometries for the specified figure IDs and save them to the specified paths.

        :param ids: List of VolumeFigure IDs in Supervisely.
        :type ids: List[int]
        :param paths: List of paths to save the downloaded geometries.
        :type paths: List[str]
        :returns: None
        :rtype: None

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                STORAGE_DIR = sly.app.get_data_dir()

                volume_id = 19371414
                project_id = 17215

                volume = api.volume.get_info_by_id(volume_id)

                key_id_map = sly.KeyIdMap()
                project_meta_json = api.project.get_meta(project_id)
                project_meta = sly.ProjectMeta.from_json(project_meta_json)

                vol_ann_json = api.volume.annotation.download(volume_id)
                id_to_paths = {}
                vol_ann = sly.VolumeAnnotation.from_json(vol_ann_json, project_meta, key_id_map)

                for sp_figure in vol_ann.spatial_figures:
                    figure_id = key_id_map.get_figure_id(sp_figure.key())
                    id_to_paths[figure_id] = f"{STORAGE_DIR}/{sp_figure.key().hex}.stl"
                if id_to_paths:
                    api.volume.figure.download_sf_geometries(*zip(*id_to_paths.items()))
        """

        if len(ids) == 0:
            return
        if len(ids) != len(paths):
            raise RuntimeError('Can not match "ids" and "paths" lists, len(ids) != len(paths)')

        id_to_path = {id: path for id, path in zip(ids, paths)}
        for figure_id, resp_part in self._download_geometries_batch(ids):
            ensure_base_path(id_to_path[figure_id])
            with open(id_to_path[figure_id], "wb") as w:
                w.write(resp_part.content)

    def load_sf_geometry(self, spatial_figure: VolumeFigure, key_id_map: KeyIdMap):
        """
        Download geometry of an existing in Supervisely figure
        and load this data into the VolumeFigure object to complete this figure representation.

        :param spatial_figure: The spatial figure object from :class:`~supervisely.volume_annotation.volume_annotation.VolumeAnnotation`.
        :type spatial_figure: :class:`~supervisely.volume_annotation.volume_figure.VolumeFigure`
        :param key_id_map: The mapped keys and IDs.
        :type key_id_map: KeyIdMap object
        """
        with tempfile.TemporaryDirectory() as temp_dir:
            figure_id = key_id_map.get_figure_id(spatial_figure.key())
            figure_path = f"{temp_dir}/{spatial_figure.key().hex}.nrrd"
            self.download_sf_geometries([figure_id], [figure_path])
            geometry = Mask3D.create_from_file(figure_path)
            spatial_figure._set_3d_geometry(geometry)

    def download(
        self, dataset_id: int, volume_ids: List[int] = None, skip_geometry: bool = False, **kwargs
    ) -> Dict[int, List[FigureInfo]]:
        """
        Method returns a dictionary with pairs of volume ID and list of FigureInfo for the given dataset ID. Can be filtered by volume IDs.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param volume_ids: Specify the list of volume IDs within the given dataset ID. If volume_ids is None, the method returns all possible pairs of images with figures. Note: Consider using `sly.batched()` to ensure that no figures are lost in the response.
        :type volume_ids: List[int], optional
        :param skip_geometry: Skip the download of figure geometry. May be useful for a significant api request speed increase in the large datasets.
        :type skip_geometry: bool

        :returns: A dictionary where keys are volume IDs and values are lists of figures.
        :rtype: Dict[int, List[:class:`~supervisely.api.entity_annotation.figure_api.FigureInfo`]]
        """
        if kwargs.get("image_ids", False) is not False:
            volume_ids = kwargs["image_ids"]  # backward compatibility
        return super().download(dataset_id, volume_ids, skip_geometry)

    def update_custom_data(
        self,
        figure_id: int,
        custom_data: Dict[str, str],
        update_strategy: Literal["replace", "merge"] = "merge",
    ) -> None:
        """
        Update custom data for a specific figure in a volume.

        :param figure_id: ID of the figure to update.
        :type figure_id: int
        :param custom_data: Custom data to update.
        :type custom_data: Dict[str, str]
        :param update_strategy: Strategy to apply, either "replace" or "merge".
        :type update_strategy: Literal["replace", "merge"]
        :returns: None
        :rtype: None

        """
        data = {
            ApiField.ID: figure_id,
            ApiField.CUSTOM_DATA: custom_data,
            ApiField.UPDATE_STRATEGY: update_strategy,
        }
        self._api.post("figures.custom-data.update", data)



import os
from collections import defaultdict
from typing import Dict, List, Optional, Union

import magic
import nrrd
import numpy as np

from supervisely import Annotation, batched, generate_free_name, is_development, ProjectMeta, logger, TagMeta, Tag, TagValueType
from supervisely.api.api import Api, ApiContext
from supervisely.convert.base_converter import AvailableImageConverters
from supervisely.convert.image.image_converter import ImageConverter
from supervisely.convert.image.medical2d import medical2d_helper as helper
from supervisely.io.fs import remove_dir, get_file_ext, mkdir, get_file_name
from supervisely.project.project_settings import LabelingInterface
from supervisely.volume.volume import is_nifti_file


# @TODO: add group tags?
class Medical2DImageConverter(ImageConverter):
    """Imports medical 2D images (NRRD, NIfTI slices) with annotations into Supervisely; supports DICOM grouping."""

    def __init__(
            self,
            input_data: str,
            labeling_interface: Optional[Union[LabelingInterface, str]],
            upload_as_links: bool,
            remote_files_map: Optional[Dict[str, str]] = None,
    ):
        """See :class:`~supervisely.convert.base_converter.BaseConverter` for params."""
        super().__init__(input_data, labeling_interface, upload_as_links, remote_files_map)

        self._filtered = None
        self._group_tag_names = defaultdict(int)

    def __str__(self):
        return AvailableImageConverters.MEDICAL2D

    def validate_labeling_interface(self) -> bool:
        """Only medical labeling interface can be used for medical images."""
        return self._labeling_interface in [
            LabelingInterface.DEFAULT,
            LabelingInterface.MEDICAL_IMAGING_SINGLE,
            LabelingInterface.IMAGES_WITH_16_COLOR,
        ]

    def validate_format(self) -> bool:
        logger.debug(f"Validating format: {self.__str__()}")

        converted_dir_name = "converted"
        converted_dir = os.path.join(self._input_data, converted_dir_name)
        mkdir(converted_dir, remove_content_if_exists=True)
        meta = ProjectMeta()

        nrrds_dict = {}
        for root, _, files in os.walk(self._input_data):
            if converted_dir_name in root:
                continue
            for file in files:
                path = os.path.join(root, file)
                ext = get_file_ext(path).lower()
                mime = magic.from_file(path, mime=True)
                if mime == "application/dicom":
                    if helper.is_dicom_file(path):  # is dicom
                        paths, names, group_tags, dcm_metas = helper.convert_dcm_to_nrrd(path, converted_dir)
                        for group_tag in group_tags:
                            if meta.get_tag_meta(group_tag["name"]) is None:
                                tag_meta = TagMeta(group_tag["name"], TagValueType.ANY_STRING)
                                meta = meta.add_tag_meta(tag_meta)
                        for path, name in zip(paths, names):
                            nrrds_dict[name] = (path, group_tags, dcm_metas)
                elif ext == ".nrrd":
                    if helper.check_nrrd(path):  # is nrrd
                        paths, names = helper.slice_nrrd_file(path, converted_dir)
                        for path, name in zip(paths, names):
                            nrrds_dict[name] = (path, None, None)
                elif mime == "application/gzip" or mime == "application/octet-stream":
                    if is_nifti_file(path):  # is nifti
                        paths, names = helper.slice_nifti_file(path, converted_dir)
                        for path, name in zip(paths, names):
                            nrrds_dict[name] = (path, None, None)

        self._items = []
        for name, (path, group_tags, dcm_metas) in nrrds_dict.items():
            item = self.Item(item_path=path)
            img_size = nrrd.read_header(path)["sizes"].tolist()[::-1] # pylint: disable=no-member
            item.set_shape(img_size)
            if group_tags is not None:
                for group_tag in group_tags:
                    self._group_tag_names[group_tag["name"]] += 1
                item.ann_data = group_tags
            item.set_meta_data(dcm_metas if dcm_metas is not None else {})
            self._items.append(item)

        self._meta = meta
        if self.items_count == 0:
            remove_dir(converted_dir)

        return self.items_count > 0

    def _get_image_channels(self, file_path: str) -> List[np.ndarray]:
        file_ext = get_file_ext(file_path).lower()
        logger.debug(f"Working with file {file_path} with extension {file_ext}.")

        if file_ext == ".nrrd":
            logger.debug(f"Found nrrd file: {file_path}.")
            image, _ = nrrd.read(file_path)
        elif file_ext == ".dcm":
            pass
        return [image[:, :, i] for i in range(image.shape[2])]

    def to_supervisely(
        self,
        item,
        meta: ProjectMeta,
        renamed_classes: dict = None,
        renamed_tags: dict = None,
    ) -> Annotation:
        ann = Annotation(item.shape)
        if item.ann_data is not None:
            tags = []
            for tag in item.ann_data:
                tag_name = renamed_tags.get(tag["name"], tag["name"])
                tag_meta = meta.get_tag_meta(tag_name)
                if tag_meta is not None:
                    group_tag = Tag(tag_meta, str(tag["value"]))
                    tags.append(group_tag)
            ann = ann.add_tags(tags)
        return ann

    def upload_dataset(
        self,
        api: Api,
        dataset_id: int,
        batch_size: int = 50,
        log_progress=True,
    ) -> None:
        """Upload converted data to Supervisely"""

        if len(self._group_tag_names) > 0:
            logger.debug("Group tags detected")
            group_tag_name = next(iter(self._group_tag_names))
            if len(self._group_tag_names) > 1:
                group_tag_name = max(self._group_tag_names, key=self._group_tag_names.get)
                logger.warn(
                    f"Multiple metadata fields found: {', '.join(self._group_tag_names.keys())}..."
                    "Some images will be hidden in the grouped view if they don't have the corresponding group tag."
                )
            meta, renamed_classes, renamed_tags = self.merge_metas_with_conflicts(api, dataset_id)
            group_tag_name = renamed_tags.get(group_tag_name, group_tag_name)
            logger.info(f"Will use [{group_tag_name}] as group tag.")

            dataset = api.dataset.get_info_by_id(dataset_id)
            existing_names = set([img.name for img in api.image.get_list(dataset_id)])

            api.project.images_grouping(id=dataset.project_id, enable=True, tag_name=group_tag_name)


            if log_progress:
                progress, progress_cb = self.get_progress(self.items_count, "Uploading images...")

            for batch in batched(self._items, batch_size):
                paths = [item.path for item in batch]
                anns = [self.to_supervisely(item, meta, renamed_classes, renamed_tags) for item in batch]
                img_metas = [item.meta or {} for item in batch]
                names = []
                for item in batch:
                    item.name = f"{get_file_name(item.path)}{get_file_ext(item.path).lower()}"
                    name = generate_free_name(
                        existing_names, item.name, with_ext=True, extend_used_names=True
                    )
                    names.append(name)

                with ApiContext(
                    api=api, project_id=dataset.project_id, dataset_id=dataset_id, project_meta=meta
                ):
                    img_infos = api.image.upload_paths(dataset_id, names, paths, metas=img_metas)
                    img_ids = [img_info.id for img_info in img_infos]
                    api.annotation.upload_anns(img_ids, anns)

                if log_progress:
                    progress_cb(len(batch))

            if log_progress:
                if is_development():
                    progress.close()
            logger.info(f"Dataset ID:'{dataset_id}' has been successfully uploaded.")
        else:
            super().upload_dataset(api, dataset_id, batch_size, log_progress)


from typing import Dict, List

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from supervisely.app import StateJson
from supervisely.app.widgets import Widget
from supervisely.api.api import Api
from supervisely.app.widgets.select_sly_utils import _get_int_or_env, _get_int_env
from supervisely.annotation.tag_meta import TagMeta, SUPPORTED_TAG_VALUE_TYPES
from supervisely.project.project_meta import ProjectMeta


class SelectTagMeta(Widget):
    """Widget to select one or multiple TagMetas from a project meta (by project ID or provided meta)."""

    class Routes:
        """Callback route names used by the widget frontend to notify Python."""

        VALUE_CHANGED = "value_changed"

    def __init__(
        self,
        default: str = None,
        project_id: int = None,
        project_meta: ProjectMeta = None,
        allowed_types: List[str] = None,
        multiselect: bool = False,
        show_label: bool = True,
        size: Literal["large", "small", "mini"] = None,
        widget_id: str = None,
    ):
        """:param default: Initial tag name.
        :type default: str, optional
        :param project_id: Project ID to load TagMetas from.
        :type project_id: int, optional
        :param project_meta: ProjectMeta with TagMetas. Mutually exclusive with project_id.
        :type project_meta: ProjectMeta, optional
        :param allowed_types: Filter by tag value types.
        :type allowed_types: List[str], optional
        :param multiselect: If True, allow multiple selection.
        :type multiselect: bool
        :param show_label: If True, show label.
        :type show_label: bool
        :param size: Size: "large", "small", or "mini".
        :type size: Literal["large", "small", "mini"], optional
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional

        :raises ValueError: If both project_id and project_meta provided, or invalid allowed_types.
        """
        self._changes_handled = False
        self._api = Api()
        self._default = default
        self._project_id = project_id
        self._project_info = None
        self._project_meta: ProjectMeta = project_meta
        if project_meta is not None and project_id is not None:
            raise ValueError(
                "You can not provide both project_id and project_meta parameters to SelectTagMeta widget."
            )

        self._allowed_types = allowed_types
        self._multiselect = multiselect
        self._show_label = show_label
        self._size = size

        if allowed_types is not None:
            for value_type in allowed_types:
                if value_type not in SUPPORTED_TAG_VALUE_TYPES:
                    raise ValueError(
                        "value_type = {!r} is unknown, should be one of {}".format(
                            value_type, SUPPORTED_TAG_VALUE_TYPES
                        )
                    )

        if project_meta is None:
            self._project_id = _get_int_or_env(self._project_id, "context.projectId")
        self._tags = None
        self._value = None
        if self._project_id is None and self._project_meta is None:
            dataset_id = _get_int_env("context.datasetId")
            if dataset_id is None:
                raise ValueError(
                    "Argument 'project_id' or environment variables 'context.projectId' or 'context.datasetId' has to be defined"
                )
            dataset_info = self._api.dataset.get_info_by_id(dataset_id, raise_error=True)
            self._project_id = dataset_info.project_id

            self._project_info = self._api.project.get_info_by_id(
                self._project_id, raise_error=True
            )

        elif self._project_meta is not None:
            self._tags = self._project_meta.tag_metas.to_json()

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self) -> Dict:
        res = {}
        res["options"] = {
            "showLabel": self._show_label,
            "filterable": True,
            "onlyAvailable": True,
            "multiple": self._multiselect,
        }
        if self._allowed_types is not None:
            res["options"]["availableValueTypes"] = self._allowed_types
        if self._size is not None:
            res["options"]["size"] = self._size
        if self._tags is not None:
            res["tags"] = self._tags
        return res

    def get_json_state(self) -> Dict:
        val = self._default
        if val is not None:
            if len(self._tags) == 0:
                val = None
            elif val not in [t["name"] for t in self._tags]:
                val = self._tags[0]["name"]
        res = {
            "projectId": self._project_id,
            "tags": val,
        }
        if self._multiselect is True:
            res["tags"] = [val]
        return res

    def get_selected_name(self) -> str:
        if self._multiselect is True:
            raise RuntimeError(
                "Tag selector allows multiselect, please use 'get_selected_names' method instead of 'get_selected_name'"
            )
        if StateJson()[self.widget_id]["tags"] == "":
            return None
        return StateJson()[self.widget_id]["tags"]

    def get_selected_names(self) -> List[str]:
        if self._multiselect is False:
            raise RuntimeError(
                "Tag selector does not allow multiselect, please use 'get_selected_name' method instead of 'get_selected_names'"
            )
        if StateJson()[self.widget_id]["tags"] is None:
            return None
        if StateJson()[self.widget_id]["tags"] == []:
            return None
        return StateJson()[self.widget_id]["tags"]

    def get_tag_meta_by_name(self, name: str) -> TagMeta:
        if self._project_id and (
            self._project_meta is None or self._project_meta.tag_metas.get(name) is None
        ):
            self._project_meta = ProjectMeta.from_json(self._api.project.get_meta(self._project_id))
        tag_meta = self._project_meta.tag_metas.get(name)
        if tag_meta is None:
            raise ValueError(f"Tag with name {name} not found in project meta")
        return tag_meta

    def get_selected_item(self) -> TagMeta:
        name = self.get_selected_name()
        if name is None:
            return None
        return self.get_tag_meta_by_name(name)

    def get_selected_items(self) -> List[TagMeta]:
        names = self.get_selected_names()
        if names is None:
            return None
        results = [self.get_tag_meta_by_name(name) for name in names]
        return results

    def set_name(self, name: str):
        if self._multiselect is True:
            self.set_names([name])
        else:
            StateJson()[self.widget_id]["tags"] = name
            StateJson().send_changes()

    def set_names(self, names: List[str]):
        if self._multiselect is False:
            raise RuntimeError(
                "Tag selector does not allow multiselect, please use 'set_name' method instead of 'set_names'"
            )
        StateJson()[self.widget_id]["tags"] = names
        StateJson().send_changes()

    def set_project_meta(self, project_meta: ProjectMeta):
        if self._project_id is not None:
            raise ValueError(
                "You can not provide both project_id and project_meta parameters to SelectTagMeta widget."
            )
        self._project_meta = project_meta
        self._tags = self._project_meta.tag_metas.to_json()
        self.update_data()
        self.update_state()
        StateJson().send_changes()

    def value_changed(self, func):
        route_path = self.get_route_path(SelectTagMeta.Routes.VALUE_CHANGED)
        server = self._sly_app.get_server()
        self._changes_handled = True

        @server.post(route_path)
        async def _click():
            if self._multiselect is True:
                res = self.get_selected_items()
            else:
                res = self.get_selected_item()
            func(res)

        return _click



import datetime
from pathlib import Path
from typing import Optional

from supervisely.api.module_api import ApiField
from supervisely.nn.benchmark.comparison.semantic_segmentation.vis_metrics import (
    Overview,
)
from supervisely.nn.benchmark.visualization.renderer import Renderer
from supervisely.nn.benchmark.visualization.widgets import (
    ContainerWidget,
    GalleryWidget,
    MarkdownWidget,
)
from supervisely.project.project_meta import ProjectMeta


class BaseComparisonVisualizer:
    """Base visualizer for model comparison reports combining multiple evaluations."""

    vis_texts = None
    ann_opacity = None
    report_name = "Model Comparison Report.lnk"

    def __init__(self, comparison):
        """:param comparison: ModelComparison instance with eval_results.
        """
        self.comparison = comparison
        self.api = comparison.api
        self.eval_results = comparison.eval_results
        self.gt_project_info = None
        self.gt_project_meta = None
        # self._widgets_created = False

        for eval_result in self.eval_results:
            eval_result.api = self.api  # add api to eval_result for overview widget
            self._get_eval_project_infos(eval_result)

        self._create_widgets()
        layout = self._create_layout()

        self.renderer = Renderer(
            layout,
            str(Path(self.comparison.workdir, "visualizations")),
            report_name=self.report_name,
        )

    def visualize(self):
        return self.renderer.visualize()

    def upload_results(self, team_id: int, remote_dir: str, progress=None):
        return self.renderer.upload_results(self.api, team_id, remote_dir, progress)

    def _create_widgets(self):
        raise NotImplementedError("Have to implement in subclasses")

    def _create_layout(self):
        raise NotImplementedError("Have to implement in subclasses")

    def _create_header(self) -> MarkdownWidget:
        """Creates header widget"""
        me = self.api.user.get_my_info().login
        current_date = datetime.datetime.now().strftime("%d %B %Y, %H:%M")
        header_main_text = " ∣ ".join(  #  vs. or | or ∣
            eval_res.name for eval_res in self.comparison.eval_results
        )
        header_text = self.vis_texts.markdown_header.format(header_main_text, me, current_date)
        header = MarkdownWidget("markdown_header", "Header", text=header_text)
        return header

    def _create_overviews(self, vm: Overview, grid_cols: Optional[int] = None) -> ContainerWidget:
        """Creates overview widgets"""
        overview_widgets = vm.overview_widgets
        if grid_cols is None:
            grid_cols = 2
            if len(overview_widgets) > 2:
                grid_cols = 3
            if len(overview_widgets) % 4 == 0:
                grid_cols = 4
        return ContainerWidget(
            overview_widgets,
            name="overview_container",
            title="Overview",
            grid=True,
            grid_cols=grid_cols,
        )

    def _create_explore_modal_table(
        self, columns_number=3, click_gallery_id=None, hover_text=None
    ) -> GalleryWidget:
        gallery = GalleryWidget(
            "all_predictions_modal_gallery",
            is_modal=True,
            columns_number=columns_number,
            click_gallery_id=click_gallery_id,
            opacity=self.ann_opacity,
        )
        gallery.set_project_meta(self.eval_results[0].pred_project_meta)
        if hover_text:
            gallery.add_image_left_header(hover_text)
        return gallery

    def _create_diff_modal_table(self, columns_number=3) -> GalleryWidget:
        gallery = GalleryWidget(
            "diff_predictions_modal_gallery",
            is_modal=True,
            columns_number=columns_number,
            opacity=self.ann_opacity,
        )
        gallery.set_project_meta(self.eval_results[0].pred_project_meta)
        return gallery

    def _create_clickable_label(self):
        return MarkdownWidget("clickable_label", "", text=self.vis_texts.clickable_label)

    def _get_eval_project_infos(self, eval_result):
        # get project infos
        if self.gt_project_info is None:
            self.gt_project_info = self.api.project.get_info_by_id(eval_result.gt_project_id)
        eval_result.gt_project_info = self.gt_project_info
        eval_result.pred_project_info = self.api.project.get_info_by_id(eval_result.pred_project_id)

        # get project metas
        if self.gt_project_meta is None:
            self.gt_project_meta = ProjectMeta.from_json(
                self.api.project.get_meta(eval_result.gt_project_id)
            )
        eval_result.gt_project_meta = self.gt_project_meta
        eval_result.pred_project_meta = ProjectMeta.from_json(
            self.api.project.get_meta(eval_result.pred_project_id)
        )

        # get dataset infos
        filters = None
        if eval_result.gt_dataset_ids is not None:
            filters = [
                {
                    ApiField.FIELD: ApiField.ID,
                    ApiField.OPERATOR: "in",
                    ApiField.VALUE: eval_result.gt_dataset_ids,
                }
            ]
        eval_result.gt_dataset_infos = self.api.dataset.get_list(
            eval_result.gt_project_id,
            filters=filters,
            recursive=True,
        )

        # eval_result.pred_dataset_infos = self.api.dataset.get_list(
        #     eval_result.pred_project_id, recursive=True
        # )



# coding: utf-8
from __future__ import annotations

import json
import uuid
from copy import deepcopy
from typing import Dict, List, Optional

from supervisely._utils import take_with_default
from supervisely.pointcloud_annotation.constants import (
    DESCRIPTION,
    FIGURES,
    KEY,
    OBJECTS,
    POINTCLOUD_ID,
    TAGS,
)
from supervisely.pointcloud_annotation.pointcloud_figure import PointcloudFigure
from supervisely.pointcloud_annotation.pointcloud_object_collection import (
    PointcloudObjectCollection,
)
from supervisely.pointcloud_annotation.pointcloud_tag_collection import (
    PointcloudTagCollection,
)
from supervisely.project.project_meta import ProjectMeta
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.video_annotation.video_annotation import VideoAnnotation


class PointcloudAnnotation(VideoAnnotation):
    """
    Annotation for a single point cloud item in Supervisely format.

    Stores point cloud-level tags, objects and figures (e.g. cuboids) and supports
    JSON (de)serialization via :meth:`to_json` / :meth:`from_json`.
    """

    def __init__(
        self,
        objects: Optional[PointcloudObjectCollection] = None,
        figures: Optional[List[PointcloudFigure]] = None,
        tags: Optional[PointcloudTagCollection] = None,
        description: Optional[str] = "",
        key: Optional[uuid.UUID] = None,
    ):
        """
        Class for creating and using PointcloudAnnotation

        :param objects: Pointcloud objects collection.
        :type objects: :class:`~supervisely.pointcloud_annotation.pointcloud_object_collection.PointcloudObjectCollection`
        :param figures: Pointcloud figures collection.
        :type figures: List[:class:`~supervisely.pointcloud_annotation.pointcloud_figure.PointcloudFigure`]
        :param tags: Pointcloud tags collection.
        :type tags: :class:`~supervisely.pointcloud_annotation.pointcloud_tag_collection.PointcloudTagCollection`
        :param description: Description text
        :type description: str
        :param key: UUID object.
        :type key: uuid.UUID

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                from supervisely.video_annotation.key_id_map import KeyIdMap

                # PointcloudAnnotation example 1
                pointcloud_ann = sly.PointcloudAnnotation()
                print(pointcloud_ann.to_json())
                # Output: {
                #     "description": "",
                #     "key": "ad97e8a4a8524b8a992d1f083c5e6b00",
                #     "tags": [],
                #     "objects": [],
                #     "figures": []
                # }

                # PointcloudAnnotation example 2
                key_id_map = KeyIdMap()
                project_meta_json = api.project.get_meta(pcd_info.project_id)
                project_meta = sly.ProjectMeta.from_json(project_meta_json)
                ann_json = api.pointcloud.annotation.download(pointcloud_id)
                ann = sly.PointcloudAnnotation.from_json(
                    data=ann_json,
                    project_meta=project_meta,
                    key_id_map=key_id_map
                )
        """
        self._description = description
        self._tags = take_with_default(tags, PointcloudTagCollection())
        self._objects = take_with_default(objects, PointcloudObjectCollection())
        self._figures = take_with_default(figures, [])
        self._key = take_with_default(key, uuid.uuid4())

    @property
    def img_size(self):
        """Not supported for pointcloud"""

        raise NotImplementedError("Not supported for pointcloud")

    @property
    def frames_count(self):
        """Not supported for pointcloud"""

        raise NotImplementedError("Not supported for pointcloud")

    @property
    def frames(self):
        """Not supported for pointcloud"""

        raise NotImplementedError("Not supported for pointcloud")

    @property
    def tags(self) -> PointcloudTagCollection:
        """
        PointcloudTag objects collection.

        :returns: Pointcloud tags collection.
        :rtype: :class:`~supervisely.pointcloud_annotation.pointcloud_tag_collection.PointcloudTagCollection`

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                project_id = 19441
                project_meta_json = api.project.get_meta(project_id)
                project_meta = sly.ProjectMeta.from_json(project_meta_json)

                # Load json file
                path = "/home/admin/work/docs/my_dataset/ann/annotation.json"
                ann = sly.PointcloudAnnotation.load_json_file(path, project_meta)

                tags = ann.tags
        """

        return super().tags

    @property
    def objects(self) -> PointcloudObjectCollection:
        """
        PointcloudObject objects collection.

        :returns: Pointcloud objects collection.
        :rtype: :class:`~supervisely.pointcloud_annotation.pointcloud_object_collection.PointcloudObjectCollection`

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                project_id = 19441
                project_meta_json = api.project.get_meta(project_id)
                project_meta = sly.ProjectMeta.from_json(project_meta_json)

                # Load json file
                path = "/home/admin/work/docs/my_dataset/ann/annotation.json"
                ann = sly.PointcloudAnnotation.load_json_file(path, project_meta)

                objects = ann.objects
        """

        return super().objects

    @property
    def figures(self) -> List[PointcloudFigure]:
        """
        PointcloudFigure objects.

        :returns: List of pointcloud figures from :class:`~supervisely.pointcloud_annotation.pointcloud_annotation.PointcloudAnnotation` object.
        :rtype: List[:class:`~supervisely.pointcloud_annotation.pointcloud_figure.PointcloudFigure`]

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                project_id = 19441
                project_meta_json = api.project.get_meta(project_id)
                project_meta = sly.ProjectMeta.from_json(project_meta_json)

                # Load json file
                path = "/home/admin/work/docs/my_dataset/ann/annotation.json"
                ann = sly.PointcloudAnnotation.load_json_file(path, project_meta)

                figures = ann.figures
        """

        return deepcopy(self._figures)

    # def get_objects_on_frame(self, frame_index: int):
    #     raise NotImplementedError("Not supported for pointcloud")

    # def get_tags_on_frame(self, frame_index: int):
    #     raise NotImplementedError("Not supported for pointcloud")

    def get_objects_from_figures(self) -> PointcloudObjectCollection:
        """
        Get PointcloudObjectCollection object from annotation figures.

        :returns: Pointcloud objects collection from annotation figures.
        :rtype: :class:`~supervisely.pointcloud_annotation.pointcloud_object_collection.PointcloudObjectCollection`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                key_id_map = KeyIdMap()
                project_id = 19441
                project_meta_json = api.project.get_meta(project_id)
                project_meta = sly.ProjectMeta.from_json(project_meta_json)
                ann_json = api.pointcloud.annotation.download(pointcloud_id)
                ann = sly.PointcloudAnnotation.from_json(
                    data=ann_json,
                    project_meta=project_meta,
                    key_id_map=key_id_map
                )

                objects = ann.get_objects_from_figures()
        """

        ann_objects = {}
        for fig in self.figures:
            if fig.parent_object.key() not in ann_objects.keys():
                ann_objects[fig.parent_object.key()] = fig.parent_object

        return PointcloudObjectCollection(ann_objects.values())

    def validate_figures_bounds(self):
        """Not supported for pointcloud"""

        raise NotImplementedError("Not supported for pointcloud")

    def to_json(self, key_id_map: Optional[KeyIdMap] = None) -> Dict:
        """
        Convert PointcloudAnnotation to json format.

        :returns: PointcloudAnnotation in json format
        :rtype: Dict

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                pointcloud_ann = sly.PointcloudAnnotation()
                print(pointcloud_ann.to_json())
                # Output: {
                #     "description": "",
                #     "key": "ad97e8a4a8524b8a992d1f083c5e6b00",
                #     "tags": [],
                #     "objects": [],
                #     "figures": []
                # }
        """

        res_json = {
            DESCRIPTION: self.description,
            KEY: self.key().hex,
            TAGS: self.tags.to_json(key_id_map),
            OBJECTS: self.objects.to_json(key_id_map),
            FIGURES: [figure.to_json(key_id_map) for figure in self.figures],
        }

        if key_id_map is not None:
            pointcloud_id = key_id_map.get_video_id(self.key())
            if pointcloud_id is not None:
                res_json[POINTCLOUD_ID] = pointcloud_id

        return res_json

    @classmethod
    def from_json(
        cls, data: Dict, project_meta: ProjectMeta, key_id_map: Optional[KeyIdMap] = None
    ) -> PointcloudAnnotation:
        """
        Convert pointcloud annotation from json format in PointcloudAnnotation object.

        :param data: Pointcloud annotation in json format.
        :type data: Dict
        :param project_meta: Project metadata.
        :type project_meta: :class:`~supervisely.project.project_meta.ProjectMeta`
        :returns: Pointcloud annotation object.
        :rtype: :class:`~supervisely.pointcloud_annotation.pointcloud_annotation.PointcloudAnnotation`

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly
                from supervisely.video_annotation.key_id_map import KeyIdMap

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                key_id_map = KeyIdMap()
                pointcloud_id = pointcloud_id
                project_id = 19441
                project_meta_json = api.project.get_meta(project_id)
                project_meta = sly.ProjectMeta.from_json(project_meta_json)
                ann_json = api.pointcloud.annotation.download(pointcloud_id)

                ann = sly.PointcloudAnnotation.from_json(
                    data=ann_json,
                    project_meta=project_meta,
                    key_id_map=key_id_map
                )
        """

        try:
            item_key = uuid.UUID(data[KEY])
        except Exception as e:
            item_key = uuid.uuid4()

        if key_id_map is not None:
            key_id_map.add_video(item_key, data.get(POINTCLOUD_ID, None))
        description = data.get(DESCRIPTION, "")
        tags = PointcloudTagCollection.from_json(data[TAGS], project_meta.tag_metas, key_id_map)
        objects = PointcloudObjectCollection.from_json(data[OBJECTS], project_meta, key_id_map)

        figures = []
        for figure_json in data.get(FIGURES, []):
            figure = PointcloudFigure.from_json(figure_json, objects, None, key_id_map)
            figures.append(figure)

        return cls(
            objects=objects, figures=figures, tags=tags, description=description, key=item_key
        )

    @classmethod
    def load_json_file(
        cls, path: str, project_meta: ProjectMeta, key_id_map: Optional[KeyIdMap] = None
    ) -> PointcloudAnnotation:
        """
        Loads json file and converts it to PointcloudAnnotation.

        :param path: Path to the json file.
        :type path: str
        :param project_meta: Input project metadata.
        :type project_meta: :class:`~supervisely.project.project_meta.ProjectMeta`
        :param key_id_map: Key ID map.
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`
        :returns: Pointcloud annotation object.
        :rtype: :class:`~supervisely.pointcloud_annotation.pointcloud_annotation.PointcloudAnnotation`

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                team_name = 'Vehicle Detection'
                workspace_name = 'Cities'
                project_name =  'London'

                team = api.team.get_info_by_name(team_name)
                workspace = api.workspace.get_info_by_name(team.id, workspace_name)
                project = api.project.get_info_by_name(workspace.id, project_name)

                project_meta_json = api.project.get_meta(project_id)
                project_meta = sly.ProjectMeta.from_json(project_meta_json)

                # Load json file
                path = "/home/admin/work/docs/my_dataset/ann/annotation.json"
                ann = sly.PointcloudAnnotation.load_json_file(path, project_meta)
        """

        with open(path) as fin:
            data = json.load(fin)
        return cls.from_json(data, project_meta, key_id_map)

    def clone(
        self,
        objects: Optional[PointcloudObjectCollection] = None,
        figures: Optional[List] = None,
        tags: Optional[PointcloudTagCollection] = None,
        description: Optional[str] = None,
    ) -> PointcloudAnnotation:
        """
        Makes a copy of PointcloudAnnotation with new fields, if fields are given, otherwise it will use fields of the original PointcloudAnnotation.

        :param objects: Pointcloud objects collection.
        :type objects: :class:`~supervisely.pointcloud_annotation.pointcloud_object_collection.PointcloudObjectCollection`
        :param figures: List of pointcloud figures.
        :type figures: List[:class:`~supervisely.pointcloud_annotation.pointcloud_figure.PointcloudFigure`]
        :param tags: Pointcloud tags collection.
        :type tags: :class:`~supervisely.pointcloud_annotation.pointcloud_tag_collection.PointcloudTagCollection`
        :param description: Description text
        :type description: str
        :returns: :class:`~supervisely.pointcloud_annotation.pointcloud_annotation.PointcloudAnnotation` class object

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly
                from supervisely.video_annotation.key_id_map import KeyIdMap

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                key_id_map = KeyIdMap()
                pointcloud_id = 19481098
                project_id = 19441
                project_meta_json = api.project.get_meta(project_id)
                project_meta = sly.ProjectMeta.from_json(project_meta_json)
                ann_json = api.pointcloud.annotation.download(pointcloud_id)
                ann = sly.PointcloudAnnotation.from_json(
                    data=ann_json,
                    project_meta=project_meta,
                    key_id_map=key_id_map
                )

                obj_class_car = sly.ObjClass('car', sly.Cuboid)
                pointcloud_obj_car = sly.PointcloudObject(obj_class_car)
                new_objects = sly.PointcloudObjectCollection([pointcloud_obj_car])

                new_ann = ann.clone(objects=new_objects)

                print(new_ann.to_json())
                # Output: {
                #     "description": "",
                #     "figures": [],
                #     "key": "2cc443272aca4cfa9c4f404614938aa7",
                #     "objects": [
                #         {
                #         "classTitle": "Pole",
                #         "createdAt": "2023-03-16T06:38:44.934Z",
                #         "key": "eff2ec5e3cda47968f45bc51b36a0dc1",
                #         "labelerLogin": "almaz",
                #         "tags": [],
                #         "updatedAt": "2023-03-16T06:38:44.934Z"
                #         },
                #         {
                #         "classTitle": "Tram",
                #         "createdAt": "2023-03-16T06:38:44.934Z",
                #         "key": "6baa92e09ceb413ba8fbfcfae74be1c7",
                #         "labelerLogin": "almaz",
                #         "tags": [],
                #         "updatedAt": "2023-03-16T06:38:44.934Z"
                #         },
                #         {
                #         "classTitle": "car",
                #         "key": "6b1bced23061437b8ddbcdd267548c96",
                #         "tags": []
                #         }
                #     ],
                #     "tags": []
                # }
        """

        return PointcloudAnnotation(
            objects=take_with_default(objects, self.objects),
            figures=take_with_default(figures, self.figures),
            tags=take_with_default(tags, self.tags),
            description=take_with_default(description, self.description),
        )



from typing import Dict, List, Union

from supervisely._utils import batched, rand_str
from supervisely.app.content import DataJson, StateJson
from supervisely.app.widgets import Container, LinePlot, Widget, generate_id
from supervisely.app.widgets.empty.empty import Empty
from supervisely.sly_logger import logger


class GridPlot(Widget):
    """Grid of LinePlot widgets; displays multiple plots in columns for comparison."""

    def __init__(
        self,
        data: List[Union[Dict, str]] = [],
        columns: int = 1,
        gap: int = 10,
        widget_id: str = None,
    ):
        """
        :param data: List of plot configs (dicts) or titles (str). Each dict is passed to LinePlot.
        :type data: List[Union[:class:`~supervisely.app.widgets.line_plot.line_plot.LinePlot`, str]]
        :param columns: Number of columns in the grid.
        :type columns: int
        :param gap: Gap between plots in pixels.
        :type gap: int
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        """
        self._widgets = {}
        self._columns = columns
        self._gap = gap

        if len(data) > 0:  # TODO: add the self.add_data([data:dict]) method
            for plot_data in data:
                if isinstance(plot_data, dict):
                    # self._widgets[plot_data['title']] = LinePlot(title=plot_data['title'], series=plot_data.get('series', []), show_legend=plot_data.get('show_legend', True))
                    # passing parameters in this way will eventually result in a JsonPatchConflict error
                    self._widgets[plot_data["title"]] = LinePlot(**plot_data)
                else:
                    self._widgets[plot_data] = LinePlot(title=plot_data, series=[])

            if self._columns < 1:
                raise ValueError(f"columns ({self._columns}) < 1")
            if self._columns > len(self._widgets):
                logger.warn(
                    f"Number of columns ({self._columns}) > number of widgets ({len(self._widgets)}). Columns are set to {len(self._widgets)}"
                )
                self._columns = len(self._widgets)

            self._content = None
            if self._columns == 1:
                self._content = Container(
                    direction="vertical",
                    widgets=self._widgets.values(),
                    gap=self._gap,
                    widget_id=generate_id(),
                )
            else:
                rows = []
                num_empty = len(self._widgets) % self._columns
                for i in range(num_empty):
                    self._widgets[generate_id()] = Empty()
                for batch in batched(list(self._widgets.values()), batch_size=self._columns):
                    rows.append(
                        Container(
                            direction="horizontal",
                            widgets=batch,
                            gap=self._gap,
                            fractions=[1] * len(batch),
                            widget_id=generate_id(),
                            overflow=None,
                        )
                    )
                self._content = Container(
                    direction="vertical",
                    widgets=rows,
                    gap=self._gap,
                    widget_id=generate_id(),
                )

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return None

    def get_json_state(self):
        return None

    def add_scalar(self, identifier: str, y, x):
        """
        Add scalar to series on plot. If no series with name,
         defined in `identifier` exists,
         one will be created automatically.

        :param identifier: slash-separated plot and series name
        :type identifier: str
        :param y: y value
        :type y: float | int
        :param x: x value
        :type x: float | int
        """
        plot_title, series_name = identifier.split("/", 1)
        _, series = self._widgets[plot_title].get_series_by_name(series_name)

        if series is not None:
            self._widgets[plot_title].add_to_series(name_or_id=series_name, data=[(x, y)])
        else:
            self._widgets[plot_title].add_series(name=series_name, x=[x], y=[y])

    def add_scalars(
        self,
        plot_title: str,
        new_values: Dict[str, Union[float, int]],
        x: Union[float, int],
    ):
        """
        Add scalars to several series on one plot at point `x`

        :param plot_title: name of existing plot
        :type plot_title: str
        :param new_values: dictionary in the `{series_name: y_value, ...}` format
        :type new_values: dict
        :param x: value of `x`
        :type x: float | int
        """
        for series_name in new_values.keys():
            _, series = self._widgets[plot_title].get_series_by_name(series_name)
            if series is not None:
                self._widgets[plot_title].add_to_series(
                    name_or_id=series_name,
                    # data=[{"x": x, "y": new_values[series_name]}],
                    data=[(x, new_values[series_name])],
                )
            else:
                self._widgets[plot_title].add_series(
                    name=series_name, x=[x], y=[new_values[series_name]]
                )



import random
import string
from os.path import join
from re import compile as re_compile
from typing import List

from supervisely.io.fs import silent_remove
from supervisely.nn.artifacts.artifacts import BaseTrainArtifacts


class MMDetection(BaseTrainArtifacts):
    """Training artifacts descriptor for MMDetection sessions (object detection / instance segmentation)."""

    def __init__(self, team_id: int):
        """See :class:`~supervisely.nn.artifacts.artifacts.BaseTrainArtifacts` for params."""
        super().__init__(team_id)

        self._app_name = "Train MMDetection"
        self._slug = "supervisely-ecosystem/mmdetection/train"
        self._serve_app_name = "Serve MMDetection"
        self._serve_slug = "supervisely-ecosystem/mmdetection/serve"
        self._framework_name = "MMDetection"
        self._framework_folder = "/mmdetection"
        self._weights_folder = "checkpoints/data"
        self._task_type = None
        self._weights_ext = ".pth"
        self._info_file = "info/ui_state.json"
        self._config_file = "config.py"
        self._pattern = re_compile(r"^/mmdetection/\d+_[^/]+/?$")
        self._available_task_types: List[str] = ["object detection", "instance segmentation"]
        self._require_runtime = False
        self._has_benchmark_evaluation = False

    def get_task_id(self, artifacts_folder: str) -> str:
        parts = artifacts_folder.split("/")
        if len(parts) < 3:
            raise ValueError(f"Invalid path: {artifacts_folder}")
        session_id, _ = parts[2].split("_", 1)
        return session_id

    def get_project_name(self, artifacts_folder: str) -> str:
        parts = artifacts_folder.split("/")
        if len(parts) < 3:
            raise ValueError(f"Invalid path: {artifacts_folder}")
        _, project_name = parts[2].split("_", 1)
        return project_name

    def get_task_type(self, artifacts_folder: str) -> str:
        info_path = join(artifacts_folder, self._info_file)
        task_type = "undefined"
        for file_info in self._get_file_infos():
            if file_info.path == info_path:
                json_data = self._fetch_json_from_path(file_info.path)
                task_type = json_data.get("task", "undefined")
                break
        return task_type

    def get_weights_path(self, artifacts_folder: str) -> str:
        return join(artifacts_folder, self._weights_folder)

    def get_config_path(self, artifacts_folder: str) -> str:
        return join(artifacts_folder, self._weights_folder, self._config_file)


class MMDetection3(BaseTrainArtifacts):
    """Training artifacts descriptor for MMDetection 3.0 sessions with benchmark evaluation support."""

    def __init__(self, team_id: int):
        """See :class:`~supervisely.nn.artifacts.artifacts.BaseTrainArtifacts` for params."""
        super().__init__(team_id)

        self._app_name = "Train MMDetection 3.0"
        self._slug = "supervisely-ecosystem/train-mmdetection-v3"
        self._serve_app_name = "Serve MMDetection 3.0"
        self._serve_slug = "supervisely-ecosystem/serve-mmdetection-v3"
        self._framework_name = "MMDetection 3.0"
        self._framework_folder = "/mmdetection-3"
        self._weights_folder = None
        self._task_type = None
        self._weights_ext = ".pth"
        self._config_file = "config.py"
        self._pattern = re_compile(r"^/mmdetection-3/\d+_[^/]+/?$")
        self._available_task_types: List[str] = ["object detection", "instance segmentation"]
        self._require_runtime = False
        self._has_benchmark_evaluation = True

    def get_task_id(self, artifacts_folder: str) -> str:
        parts = artifacts_folder.split("/")
        if len(parts) < 3:
            raise ValueError(f"Invalid path: {artifacts_folder}")
        session_id, _ = parts[2].split("_", 1)
        return session_id

    def get_project_name(self, artifacts_folder: str) -> str:
        config_path = join(artifacts_folder, self._config_file)
        config_name = "".join(random.choices(string.ascii_lowercase + string.digits, k=10)) + ".txt"
        self._api.file.download(self._team_id, config_path, config_name)
        project_name = None
        with open(config_name, "r") as f:
            lines = f.readlines()
            project_line = lines[-1] if lines else None
            if project_line is None:
                f.close()
                silent_remove(config_name)
                return project_name
            start = project_line.find("'") + 1
            end = project_line.find("'", start)
            project_name = project_line[start:end]
            f.close()
        silent_remove(config_name)
        return project_name

    def get_task_type(self, artifacts_folder: str) -> str:
        config_path = join(artifacts_folder, self._config_file)
        config_name = "".join(random.choices(string.ascii_lowercase + string.digits, k=10)) + ".txt"
        self._api.file.download(self._team_id, config_path, config_name)
        task_type = "undefined"
        with open(config_name, "r") as f:
            lines = f.readlines()
            task_type_line = lines[-3] if lines else None
            if task_type_line is None:
                f.close()
                silent_remove(config_name)
                return task_type
            start = task_type_line.find("'") + 1
            end = task_type_line.find("'", start)
            task_type = task_type_line[start:end].replace("_", " ")
            f.close()
        silent_remove(config_name)
        return task_type

    def get_weights_path(self, artifacts_folder: str) -> str:
        return artifacts_folder

    def get_config_path(self, artifacts_folder: str) -> str:
        return join(artifacts_folder, self._config_file)



import os
import json
import random
from typing import List, Optional, Dict
from collections import namedtuple

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

import supervisely as sly
from supervisely.app import StateJson, DataJson
from supervisely.app.widgets import Container, Widget, RadioTabs, Editor, GridGallery, Select, Input, Field, Button


class Augmentations(Widget):
    """Widget to load an imgaug augmentation pipeline and preview its effect on a random image."""

    def __init__(
        self,
        globals,
        image_info: sly.ImageInfo = None, 
        task_type: Optional[Literal['detection', 'segmentation']] = None,
        remote_preview_path: str = '/temp/preview_augs.jpg',
        widget_id=None
    ):
        """
        :param globals: Globals object with api, project_meta, data_dir, team.
        :type globals: object
        :param image_info: Optional image for preview. If None, random image is used.
        :type image_info: sly.ImageInfo, optional
        :param task_type: Task type for annotation conversion: "detection" or "segmentation".
        :type task_type: Literal["detection", "segmentation"], optional
        :param remote_preview_path: Path in Team Files for uploaded preview.
        :type remote_preview_path: str
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        """
        self._globals = globals
        self._image_info = image_info
        self._task_type = task_type
        self._remote_preview_path = remote_preview_path

        self._template_path  = None
        self._pipeline = None
        self._py_code = None

        self._editor = Editor(
            language_mode='python', 
            height_px=250, 
            readonly=True, 
            show_line_numbers=True,
            highlight_active_line=False,
            restore_default_button=False,
        )
        self._grid_gallery = GridGallery(
            columns_number=2,
            annotations_opacity=0.3,
            show_opacity_slider=True,
            enable_zoom=True,
            resize_on_zoom=False,
            sync_views=True,
            fill_rectangle=False,
        )
        self._grid_gallery.hide()
        self._button_preview = Button('Preview on random image')
        self._content = Container([self._editor, self._grid_gallery, self._button_preview])

        @self._button_preview.click
        def update_preview():
            self.preview_augs()
        super().__init__(widget_id=widget_id, file_path=__file__)

    def preview_augs(self, image_info: sly.ImageInfo = None):
        if not image_info:
            # ds_name, item_name = self.get_random_item()
            # self._image_info = self.get_image_info_from_cache(ds_name, item_name)
            self._image_info = self.get_random_image_info()

        img = self._globals.api.image.download_np(self._image_info.id)
        ann_json = self._globals.api.annotation.download(self._image_info.id).annotation

        image_ann = sly.Annotation.from_json(ann_json, self._globals.project_meta)
        meta = self._globals.project_meta
        if self._task_type == "detection":
            image_ann, meta = self.convert_ann_to_bboxes(image_ann, self._globals.project_meta)

        _, res_img, res_ann = sly.imgaug_utils.apply(self._pipeline, meta, img, image_ann)
        local_image_path = os.path.join(self._globals.data_dir, "preview_augs.jpg")
        sly.image.write(local_image_path, res_img)
        if self._globals.api.file.exists(self._globals.team.id, self._remote_preview_path):
            self._globals.api.file.remove(self._globals.team.id, self._remote_preview_path)
        file_info = self._globals.api.file.upload(self._globals.team.id, local_image_path, self._remote_preview_path)

        self._grid_gallery.clean_up()
        self._grid_gallery.append(
            title=f"Original", image_url=self._image_info.full_storage_url, annotation=image_ann
        )
        self._grid_gallery.append(
            title=f"Augmented", image_url=file_info.full_storage_url, annotation=res_ann
        )
        self._grid_gallery.show()

    def get_random_image_info(self):
        api: sly.Api = self._globals.api
        project_id = sly.env.project_id(False) or self._globals.PROJECT_ID
        datasets = api.dataset.get_list(project_id)
        r_dataset = random.choice(datasets)
        img_infos = api.image.get_list(r_dataset.id)
        return random.choice(img_infos)

    # def get_random_item(self):
    #     all_ds_names = self._globals.project_fs.datasets.keys()
    #     ds_name = random.choice(all_ds_names)
    #     ds = self._globals.project_fs.datasets.get(ds_name)
    #     items = list(ds)
    #     item_name = random.choice(items)
    #     return ds_name, item_name

    # def get_image_info_from_cache(self, dataset_name, item_name):
    #     dataset_fs = self._globals.project_fs.datasets.get(dataset_name)
    #     img_info_path = dataset_fs.get_img_info_path(item_name)
    #     image_info_dict = sly.json.load_json_file(img_info_path)
    #     ImageInfo = namedtuple('ImageInfo', image_info_dict)
    #     info = ImageInfo(**image_info_dict)
    #     return info

    def update_augmentations(self, path_or_data: str = None, string_format: str = 'python'):
        if path_or_data.endswith('.json'):
            self._template_path = path_or_data
            config = sly.json.load_json_file(self._template_path)
            self._pipeline, self._py_code = self.load_augs_template(config)
        elif path_or_data.endswith('.py'):
            # TODO add support .py files
            raise NotImplementedError('.py files not supported yet.')
        else:
            if string_format == 'json':
                config = json.loads(path_or_data)
                self._pipeline, self._py_code = self.load_augs_template(config)
            elif string_format == 'python':
                # TODO create func to conversion augs
                raise NotImplementedError('Raw python augmentations not supported yet.')
            else:
                raise ValueError('Supported values for "string_format" is "python" or "json"')
        self._editor.set_text(text=self._py_code)

    def get_augmentations(self):
        return self._pipeline, self._py_code

    def get_json_data(self) -> Dict:
        return {}

    def get_json_state(self) -> Dict:
        return {}

    @staticmethod
    def convert_ann_to_bboxes(image_ann, project_meta):
        meta = project_meta.clone()
        for obj_class in meta.obj_classes:
            if obj_class.geometry_type == "rectangle":
                continue
            class_obj = sly.ObjClass(obj_class.name, sly.Rectangle, obj_class.color)
            meta = meta.delete_obj_class(obj_class.name)
            meta = meta.add_obj_class(class_obj)
        new_ann_json = {
            "size": {
                "height": image_ann.img_size[0],
                "width": image_ann.img_size[1]
            },
            "tags": [],
            "objects": []
        }
        new_ann = sly.Annotation.from_json(new_ann_json, meta)
        for label in image_ann.labels:
            if label.geometry.geometry_name == "rectangle":
                new_ann = image_ann.add_label(label)
                continue
            class_obj = sly.ObjClass(label.obj_class.name, sly.Rectangle, label.obj_class.color)
            updated_label = label.convert(class_obj)[0]
            new_ann = new_ann.add_label(updated_label)
        return new_ann, meta

    @staticmethod
    def load_augs_template(config: dict):
        pipeline = sly.imgaug_utils.build_pipeline(config["pipeline"], random_order=config["random_order"]) # to validate
        py_code = sly.imgaug_utils.pipeline_to_python(config["pipeline"], config["random_order"])
        return pipeline, py_code

    def disable(self):
        self._disabled = True
        self._editor.disable()
        self._grid_gallery.disable()
        self._button_preview.disable()
        DataJson()[self.widget_id]["disabled"] = self._disabled
        DataJson().send_changes()

    def enable(self):
        self._disabled = False
        self._editor.enable()
        self._grid_gallery.enable()
        self._button_preview.enable()
        DataJson()[self.widget_id]["disabled"] = self._disabled
        DataJson().send_changes()


class AugmentationsWithTabs(Widget):
    """Augmentations widget with tabs for choosing a template or loading a custom pipeline from Team Files."""

    def __init__(
        self,
        globals,
        templates: List[Dict[str, str]],
        image_info: sly.ImageInfo = None, 
        task_type: Optional[Literal['detection', 'segmentation']] = None,
        remote_preview_path: str = '/temp/preview_augs.jpg',
        widget_id=None
    ):
        """
        :param globals: Globals object with api, project_meta, data_dir, team.
        :type globals: object
        :param templates: List of dicts with 'value' and 'label' for template selector.
        :type templates: List[Dict[str, str]]
        :param image_info: Optional image for preview.
        :type image_info: sly.ImageInfo, optional
        :param task_type: Task type: "detection" or "segmentation".
        :type task_type: Literal["detection", "segmentation"], optional
        :param remote_preview_path: Path in Team Files for preview upload.
        :type remote_preview_path: str
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        """
        self._globals = globals
        self._image_info = image_info
        self._templates = templates
        self._task_type = task_type
        self._remote_preview_path = remote_preview_path

        self._augs1 = Augmentations(globals, image_info=self._image_info, task_type=self._task_type)
        self._augs2 = Augmentations(globals, image_info=self._image_info, task_type=self._task_type)
        self._augs2._editor.hide()
        self._augs2._button_preview.hide()
        self._template_path_input = Input(placeholder="Path to .json file in Team Files")
        self._template_selector = Select(
            items=[Select.Item(value=t['value'], label=t['label']) for t in templates], 
            filterable=True, 
            placeholder="select me"
        )
        self._button_template_update = Button('Load template from file')
        self._radio_tabs = RadioTabs(
            titles=["From template", "Custom pipeline"],
            contents=[
                Field(
                    title='Template', 
                    content=Container([self._template_selector, self._augs1]),
                ),
                Field(
                    title='Path to JSON configuration', 
                    description='Copy path in Team Files', 
                    content=Container([self._template_path_input, self._button_template_update, self._augs2])
                )
            ],
            descriptions=[
                "Choose one of the prepared templates",
                "Use ImgAug Studio appto configure and save custom augmentations",
            ],
        )

        self._current_augs = self._augs1
        self._current_augs.update_augmentations(templates[0]['value'])

        @self._radio_tabs.value_changed
        def tab_toggle(tab_title):
            if tab_title == 'Custom pipeline':
                self._current_augs = self._augs2
            else:
                self._current_augs = self._augs1

        @self._template_selector.value_changed
        def selector_value_changed(value = None):
            self._current_augs.update_augmentations(value)

        @self._button_template_update.click
        def update_template():
            custom_template_path = self._template_path_input.get_value()
            if self._globals.api.file.exists(team_id=self._globals.team.id, remote_path=custom_template_path.strip()):
                self.load_existing_pipeline(custom_template_path)
                self._augs2._editor.show()
                self._augs2._button_preview.show()
            else:
                raise FileExistsError("Team files does't contains file by passed path.")

        self._content = Container([self._radio_tabs])
        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self) -> Dict:
        return {}

    def get_json_state(self) -> Dict:
        return {}

    def get_augmentations(self):
        return self._current_augs.get_augmentations()

    def load_existing_pipeline(self, remote_path):
        custom_pipeline_path = os.path.join(self._globals.data_dir, sly.fs.get_file_name_with_ext(remote_path))
        self._globals.api.file.download(self._globals.team.id, remote_path, custom_pipeline_path)
        self._current_augs.update_augmentations(custom_pipeline_path)

    def disable(self):
        self._disabled = True
        self._radio_tabs.disable()
        self._current_augs.disable()
        self._template_selector.disable()
        self._template_path_input.disable()
        self._button_template_update.disable()
        DataJson()[self.widget_id]["disabled"] = self._disabled
        DataJson().send_changes()

    def enable(self):
        self._disabled = False
        self._radio_tabs.enable()
        self._current_augs.enable()
        self._template_selector.enable()
        self._template_path_input.enable()
        self._button_template_update.enable()
        DataJson()[self.widget_id]["disabled"] = self._disabled
        DataJson().send_changes()



from __future__ import annotations

import json
import os
import pickle
import zipfile
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import numpy as np
import pandas as pd
import yaml

from supervisely.app.widgets import SlyTqdm
from supervisely.io.fs import get_file_name_with_ext, silent_remove
from supervisely.io.json import dump_json_file, load_json_file
from supervisely.task.progress import tqdm_sly


class BaseEvalResult:
    """Base container for evaluation outputs loaded from an evaluator result directory."""

    PRIMARY_METRIC = None

    def __init__(self, directory: Optional[str] = None):
        """:param directory: Path to evaluator result directory.
        """
        self.directory = directory
        self.inference_info: Dict = {}
        self.speedtest_info: Optional[Dict] = None
        self.eval_data: Optional[Dict] = None
        self.mp = None

        if self.directory is not None:
            self._read_files(self.directory)
            self._prepare_data()

    @classmethod
    def from_evaluator(cls, evaulator: BaseEvaluator) -> BaseEvalResult:
        """Method to customize loading of the evaluation result."""
        raise NotImplementedError()

    @property
    def cv_task(self):
        return self.inference_info.get("task_type")

    @property
    def name(self) -> Union[str, None]:
        deploy_params = self.inference_info.get("deploy_params", {})
        return (
            deploy_params.get("checkpoint_name")
            or deploy_params.get("model_name")
            or self.inference_info.get("model_name")
        )

    @property
    def short_name(self) -> Union[str, None]:
        if not self.name:
            return None
        if len(self.name) > 20:
            return self.name[:9] + "..." + self.name[-7:]
        return self.name

    @property
    def gt_project_id(self) -> Optional[int]:
        return self.inference_info.get("gt_project_id")

    @property
    def gt_dataset_ids(self) -> Optional[List[int]]:
        return self.inference_info.get("gt_dataset_ids", None)

    @property
    def dt_project_id(self):
        return self.inference_info.get("dt_project_id")

    @property
    def pred_project_id(self):
        return self.dt_project_id

    @property
    def train_info(self):
        return self.inference_info.get("train_info", None)  # TODO: check

    @property
    def evaluator_app_info(self):
        return self.inference_info.get("evaluator_app_info", None)  # TODO: check

    @property
    def val_images_cnt(self):
        return self.inference_info.get("val_images_cnt", None)  # TODO: check

    @property
    def classes_whitelist(self):
        return self.inference_info.get("inference_settings", {}).get("classes", [])  # TODO: check

    def _read_files(self, path: str) -> None:
        """Read all necessary files from the directory"""
        raise NotImplementedError()

    def _prepare_data(self) -> None:
        """Prepare data to allow easy access to the data"""
        raise NotImplementedError()

    @property
    def key_metrics(self):
        raise NotImplementedError()

    @property
    def checkpoint_name(self):
        if self.inference_info is None:
            return None

        deploy_params = self.inference_info.get("deploy_params", {})
        name = None
        if deploy_params:
            name = deploy_params.get("checkpoint_name")  # not TrainApp
            if name is None:
                name = deploy_params.get("model_files", {}).get("checkpoint")
                if name is not None:
                    name = get_file_name_with_ext(name)
        if name is None:
            name = self.inference_info.get("checkpoint_name", "")
        return name

    def _load_eval_data_archive(self, path: Path, pd_index_col=False) -> Dict:
        """Load eval_data from archive"""
        with zipfile.ZipFile(path, mode="r") as zf:
            with zf.open("eval_data.json") as json_f:
                data = json.load(json_f)
            return self._process_value_from_archive(data, zf, pd_index_col)

    def _process_value_from_archive(self, value, zf: zipfile.ZipFile, pd_index_col: bool = False):
        """Recursively process values from archive, handling nested dicts and lists."""
        if isinstance(value, str) and value.endswith(".npy"):
            with zf.open(value) as arr_f:
                return np.load(arr_f)
        elif isinstance(value, str) and value.endswith(".csv"):
            with zf.open(value) as df_f:
                if pd_index_col:
                    return pd.read_csv(df_f, sep="\t", index_col=0)
                return pd.read_csv(df_f, sep="\t")
        elif isinstance(value, dict):
            res = {}
            for k, v in value.items():
                k = int(k) if isinstance(k, str) and k.isdigit() else k
                k = float(k) if isinstance(k, str) and self._is_float(k) else k
                res[k] = self._process_value_from_archive(v, zf, pd_index_col)
            return res
        elif isinstance(value, list):
            return [self._process_value_from_archive(item, zf, pd_index_col) for item in value]
        elif isinstance(value, str) and value.isdigit():
            return int(value)
        else:
            return value

    def _is_float(self, s: str) -> bool:
        if not s or not isinstance(s, str):
            return False
        try:
            float(s)
            return "." in s or "e" in s.lower()
        except (ValueError, AttributeError):
            return False


class BaseEvaluator:
    """Base class for benchmark evaluators that compare ground truth and prediction projects."""

    EVALUATION_PARAMS_YAML_PATH: Optional[str] = None
    eval_result_cls = BaseEvalResult

    def __init__(
        self,
        gt_project_path: str,
        pred_project_path: str,
        result_dir: str = "./evaluation",
        progress: Optional[SlyTqdm] = None,
        items_count: Optional[int] = None,  # TODO: is it needed?
        classes_whitelist: Optional[List[str]] = None,
        evaluation_params: Optional[dict] = None,
    ):
        """:param gt_project_path: Path to ground-truth project.
        :param pred_project_path: Path to predictions project.
        :param result_dir: Output directory for evaluation results.
        :param progress: Progress callback.
        :param items_count: Optional total items count.
        :param classes_whitelist: Optional class filter.
        :param evaluation_params: Optional evaluation config.
        """
        self.eval_data: Optional[Dict] = None
        self.gt_project_path = gt_project_path
        self.pred_project_path = pred_project_path
        self.result_dir = result_dir
        self.total_items = items_count
        self.pbar = progress or tqdm_sly
        os.makedirs(result_dir, exist_ok=True)
        self.classes_whitelist = classes_whitelist or []

        if evaluation_params is None:
            evaluation_params = self._get_default_evaluation_params()
        self.evaluation_params = evaluation_params
        if self.evaluation_params:
            self.validate_evaluation_params(self.evaluation_params)

    def evaluate(self):
        raise NotImplementedError()

    @classmethod
    def load_yaml_evaluation_params(cls) -> Union[str, None]:
        if cls.EVALUATION_PARAMS_YAML_PATH is None:
            return None
        with open(cls.EVALUATION_PARAMS_YAML_PATH, "r") as f:
            return f.read()

    @classmethod
    def validate_evaluation_params(cls, evaluation_params: dict) -> None:
        pass

    @classmethod
    def _get_default_evaluation_params(cls) -> dict:
        if cls.EVALUATION_PARAMS_YAML_PATH is None:
            return {}
        else:
            params = cls.load_yaml_evaluation_params()
            if params is None:
                return {}
            return yaml.safe_load(params)

    def _dump_pickle(self, data, file_path):
        with open(file_path, "wb") as f:
            pickle.dump(data, f)

    def _process_value_for_archive(self, value: Any, key_prefix: str, zf: zipfile.ZipFile) -> Any:
        """Recursively process values for archiving, handling nested dicts and lists."""
        if isinstance(value, np.ndarray):
            filename = f"{key_prefix}.npy" if key_prefix else "array.npy"
            filepath = os.path.join(self.result_dir, filename)
            np.save(filepath, value)
            zf.write(filepath, arcname=filename)
            os.remove(filepath)
            return filename
        elif isinstance(value, pd.DataFrame):
            filename = f"{key_prefix}.csv" if key_prefix else "dataframe.csv"
            filepath = os.path.join(self.result_dir, filename)
            value.to_csv(filepath, sep="\t")
            zf.write(filepath, arcname=filename)
            os.remove(filepath)
            return filename
        elif isinstance(value, dict):
            return {
                k: self._process_value_for_archive(v, f"{key_prefix}.{k}" if key_prefix else k, zf)
                for k, v in value.items()
            }
        elif isinstance(value, list):
            return [
                self._process_value_for_archive(item, f"{key_prefix}[{i}]", zf)
                for i, item in enumerate(value)
            ]
        elif isinstance(value, (np.integer, np.floating)):
            return value.item()
        elif isinstance(value, np.bool_):
            return bool(value)
        elif isinstance(value, str) and value.isdigit():
            return int(value)
        else:
            return value

    def _dump_eval_results_archive(self):
        with zipfile.ZipFile(os.path.join(self.result_dir, "eval_data.zip"), mode="w") as zf:
            data = self._process_value_for_archive(self.eval_data, "", zf)
            filepath = os.path.join(self.result_dir, "eval_data.json")
            dump_json_file(data, filepath, indent=4)
            zf.write(filepath, arcname="eval_data.json")
            silent_remove(filepath)

    def get_eval_result(self) -> BaseEvalResult:
        return self.eval_result_cls(self.result_dir)



from __future__ import annotations

from typing import Any, Dict, Iterator, List, Optional

import supervisely.sly_logger as logger
from supervisely.annotation.tag_collection import TagCollection
from supervisely.annotation.tag_meta import TagMeta
from supervisely.annotation.tag_meta_collection import TagMetaCollection
from supervisely.pointcloud_annotation.pointcloud_tag import PointcloudTag
from supervisely.video_annotation.key_id_map import KeyIdMap


class PointcloudTagCollection(TagCollection):
    """
    Collection with :class:`~supervisely.pointcloud_annotation.pointcloud_tag.PointcloudTag` instances.
    :class:`~supervisely.pointcloud_annotation.pointcloud_tag_collection.PointcloudTagCollection` object is immutable.

    :Usage Example:

        .. code-block:: python

            import supervisely as sly

            # Create TagMetas (see class TagMeta for additional information about creating TagMetas)
            meta_weather = sly.TagMeta('Weather', sly.TagValueType.ANY_STRING)

            seasons = ["Winter", "Spring", "Summer", "Autumn"]
            meta_season = sly.TagMeta('Season', sly.TagValueType.ONEOF_STRING, possible_values=seasons)

            # Create Pointcloud Tags
            tag_weather = sly.PointcloudTag(meta_weather, value="Sunny")
            tag_season = sly.PointcloudTag(meta_season, value="Spring")
            tags_arr = [tag_weather, tag_season]

            # Create PointcloudTagCollection from Pointcloud Tags
            tags = sly.PointcloudTagCollection(tags_arr)

            # Add item to PointcloudTagCollection
            meta_potato = sly.TagMeta('potato', sly.TagValueType.NONE)
            tag_potato = sly.PointcloudTag(meta_potato)

            # Remember that PointcloudTagCollection is immutable, and we need to assign new instance of PointcloudTagCollection to a new variable
            tags = tags.add(tag_potato)

            # You can also add multiple items to collection
            meta_cabbage = sly.TagMeta('cabbage', sly.TagValueType.NONE)
            meta_carrot = sly.TagMeta('carrot', sly.TagValueType.NONE)
            meta_turnip = sly.TagMeta('turnip', sly.TagValueType.NONE)

            tag_cabbage = sly.PointcloudTag(meta_cabbage)
            tag_carrot = sly.PointcloudTag(meta_carrot)
            tag_turnip = sly.PointcloudTag(meta_turnip)

            additional_veggies = [tag_cabbage, tag_carrot, tag_turnip]

            tags = tags.add_items(additional_veggies)

            # Has key, checks if given key exist in collection
            tags.has_key(tag_cabbage.key())
            # Output: True

            # Intersection, finds intersection of given list of instances with collection items
            meta_dog = sly.TagMeta('dog', sly.TagValueType.NONE)
            tag_dog = sly.PointcloudTag(meta_dog)

            meta_cat = sly.TagMeta('cat', sly.TagValueType.NONE)
            tag_cat = sly.PointcloudTag(meta_cat)

            meta_turtle = sly.TagMeta('turtle', sly.TagValueType.NONE)
            tag_turtle = sly.PointcloudTag(meta_turtle)

            tags_animals = sly.PointcloudTagCollection([tag_dog, tag_cat, tag_turtle])

            tags_intersections = tags.intersection(tags_animals)
            print(tags_intersections.to_json())
            # Output: []

            # Let's add the potato Tag from another collection and compare them again
            tags_animals = tags_animals.add(tag_potato)

            tags_intersections = tags.intersection(tags_animals)
            print(tags_intersections.to_json())
            # Output: [
            #     {
            #         "name":"potato",
            #         "key": "058ad7993a534082b4d94cc52542a97d"
            #     }
            # ]

            # Difference, finds difference between collection and given list of Tags or TagCollection
            meta_car = sly.TagMeta('car', sly.TagValueType.NONE)
            tag_car = sly.Tag(meta_car)

            meta_bicycle = sly.TagMeta('bicycle', sly.TagValueType.NONE)
            tag_bicycle = sly.PointcloudTag(meta_bicycle)

            tags_vehicles = sly.PointcloudTagCollection([tag_car, tag_bicycle])

            meta_pedestrian = sly.TagMeta('pedestrian', sly.TagValueType.NONE)
            tag_pedestrian = sly.PointcloudTag(meta_pedestrian)

            meta_road = sly.TagMeta('road', sly.TagValueType.NONE)
            tag_road = sly.PointcloudTag(meta_road)

            difference = tags_vehicles.difference([tag_pedestrian, tag_road])
            print(difference.to_json())
            # Output: [
            #     {
            #         "name": "car",
            #         "key": "058ad7993a534082b4d94cc52542a97d"
            #     },
            #     {
            #         "name": "bicycle",
            #         "key": ""0c0033c5b4834d4cbabece4317295f07"
            #     }
            # ]

            # Merge, merges collection and given list of TagMetas
            tags_vehicles = sly.TagMetaCollection([tag_car, tag_bicycle])
            tags_merge = sly.TagMetaCollection([tag_pedestrian, tag_road])

            merged_collections = tags_vehicles.merge(tags_merge)
            print(merged_collections.to_json())
            # Output: [
            #     {
            #         "name":"pedestrian"
            #     },
            #     {
            #         "name":"road"
            #     },
            #     {
            #         "name":"car"
            #     },
            #     {
            #         "name":"bicycle"
            #     }
            # ]
    """

    item_type = PointcloudTag

    def __iter__(self) -> Iterator[PointcloudTag]:
        return next(self)

    # @classmethod
    # def from_api_response(
    #     cls,
    #     data: List[Dict],
    #     tag_meta_collection: TagMetaCollection,
    #     id_to_tagmeta: Optional[Dict[int, TagMeta]] = None
    # ) -> PointcloudTagCollection:
    #     return super().from_api_response(data, tag_meta_collection, id_to_tagmeta)

    # def get_by_name(self, tag_name: str) -> List[PointcloudTag]:
    #     """
    #     Get list of Pointcloud Tags with provided name.

    #     :param tag_name: Pointcloud Tag name.
    #     :type tag_name: str
    #     :returns: List of Pointcloud Tags.
    #     :rtype: List[:class:`~supervisely.pointcloud_annotation.pointcloud_tag.PointcloudTag`]
    #     """
    #     res = []
    #     for tag in self:
    #         if tag.name == tag_name:
    #             res.append(tag)
    #     return res

    # def get_single_by_name(self, tag_name: str) -> PointcloudTag:
    #     """
    #     Get Pointcloud Tag with provided name.
    #     Method will raise error If collection contains more than 1 tag with provided name.

    #     :param tag_name: Pointcloud Tag name.
    #     :type tag_name: str
    #     :returns: PointcloudTag object or None If no elements with provided name in collection.
    #     :rtype: :class:`~supervisely.pointcloud_annotation.pointcloud_tag.PointcloudTag` or :class:`NoneType`
    #     :raises ValueError: If collection contains more than 1 tag with provided name.
    #     """
    #     res = []
    #     for tag in self:
    #         if tag.name == tag_name:
    #             res.append(tag)
    #     if len(res) == 0:
    #         return None
    #     if len(res) > 1:
    #         raise ValueError(
    #             f"There are more than one tag {tag_name} in VideoTagCollection. Use method get_by_name instead"
    #         )
    #     return res[0]

    def to_json(self, key_id_map: Optional[KeyIdMap] = None) -> List[Dict]:
        """
        Convert the :class:`~supervisely.pointcloud_annotation.pointcloud_tag_collection.PointcloudTagCollection` to a list of json dicts.
        Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :param key_id_map: Key ID map.
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`
        :returns: List of dictionaries in json format.
        :rtype: List[Dict]

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                meta_weather = sly.TagMeta('Weather', sly.TagValueType.ANY_STRING)

                seasons = ["Winter", "Spring", "Summer", "Autumn"]
                meta_season = sly.TagMeta('Season', sly.TagValueType.ONEOF_STRING, possible_values=seasons)

                tag_weather = sly.PointcloudTag(meta_weather, value="Sunny")
                tag_season = sly.PointcloudTag(meta_season, value="Spring")

                tags_arr = [tag_weather, tag_season]
                tags = sly.PointcloudTagCollection(tags_arr)

                tags_json = tags.to_json()
                # Output: [
                #     {
                #         "name":"Weather",
                #         "value":"Sunny",
                #         "key": "058ad7993a534082b4d94cc52542a97d"
                #     },
                #     {
                #         "name":"Season",
                #         "value":"Spring",
                #         "key": "0c0033c5b4834d4cbabece4317295f07"
                #     }
                # ]
        """

        return [tag.to_json(key_id_map) for tag in self]

    @classmethod
    def from_json(
        cls,
        data: List[Dict],
        tag_meta_collection: TagMetaCollection,
        key_id_map: Optional[KeyIdMap] = None,
    ) -> PointcloudTagCollection:
        """
        Convert a list with dictionaries in json format to :class:`~supervisely.pointcloud_annotation.pointcloud_tag_collection.PointcloudTagCollection`.
        Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :param data: List with dicts in json format.
        :type data: List[Dict]
        :param tag_meta_collection: Tag metadata collection.
        :type tag_meta_collection: :class:`~supervisely.annotation.tag_meta_collection.TagMetaCollection`
        :param key_id_map: Key ID map.
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`
        :returns: Pointcloud tag collection.
        :rtype: :class:`~supervisely.pointcloud_annotation.pointcloud_tag_collection.PointcloudTagCollection`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                # Initialize TagMetaCollection
                meta_weather = sly.TagMeta('Weather', sly.TagValueType.ANY_STRING)

                seasons = ["Winter", "Spring", "Summer", "Autumn"]
                meta_season = sly.TagMeta('Season', sly.TagValueType.ONEOF_STRING, possible_values=seasons)

                metas_arr = [meta_weather, meta_season]
                tag_metas = sly.TagMetaCollection(metas_arr)

                data = [
                    {"name":"Weather", "value":"Sunny"},
                    {"name":"Season", "value":"Spring"}
                ]

                tags = sly.PointcloudTagCollection.from_json(data, tag_metas)
        """

        tags = [
            cls.item_type.from_json(tag_json, tag_meta_collection, key_id_map) for tag_json in data
        ]
        return cls(tags)



# coding: utf-8

"""Work with point cloud figures via the Supervisely API."""

from typing import Dict, List, Optional

from supervisely.api.module_api import ApiField
from supervisely.api.entity_annotation.figure_api import FigureApi, FigureInfo
from supervisely.pointcloud_annotation.pointcloud_figure import PointcloudFigure
from supervisely.video_annotation.key_id_map import KeyIdMap


class PointcloudFigureApi(FigureApi):
    """
    API for working with :class:`~supervisely.pointcloud_annotation.pointcloud_figure.PointcloudFigure`.
    :class:`~supervisely.api.pointcloud.pointcloud_figure_api.PointcloudFigureApi` object is immutable.
    """

    def create(
        self,
        pointcloud_id: int,
        object_id: int,
        geometry_json: Dict,
        geometry_type: str,
        track_id: Optional[int] = None,
    ) -> int:
        """
        Create new PointcloudFigure of given point cloud object in point cloud with given ID.

        :param pointcloud_id: Point cloud ID in Supervisely.
        :type pointcloud_id: int
        :param object_id: ID of the object to which the PointcloudFigure belongs.
        :type object_id: int
        :param geometry_json: Parameters of geometry for :class:`~supervisely.pointcloud_annotation.pointcloud_figure.PointcloudFigure`.
        :type geometry_json: dict
        :param geometry_type: Type of PointcloudFigure geometry.
        :type geometry_type: str
        :param track_id: int, optional.
        :type track_id: int, optional
        :returns: New figure ID
        :rtype: int

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                pcd_id = 19618685
                object_id = 5565921
                geometry_json = {'points': {'exterior': [[500, 500], [1555, 1500]], 'interior': []}}
                geometry_type = 'rectangle'

                figure_id = api.pointcloud.figure.create(pcd_id, object_id, geometry_json, geometry_type) # 643182610
        """

        return super().create(pointcloud_id, object_id, {}, geometry_json, geometry_type, track_id)

    def append_bulk(
        self,
        pointcloud_id: int,
        figures: List[PointcloudFigure],
        key_id_map: KeyIdMap,
    ) -> None:
        """
        Add PointcloudFigures to given point cloud by ID.

        :param pointcloud_id: Point cloud ID in Supervisely.
        :type pointcloud_id: int
        :param figures: List of point cloud figures to append.
        :type figures: List[:class:`~supervisely.pointcloud_annotation.pointcloud_figure.PointcloudFigure`]
        :param key_id_map: KeyIdMap object.
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`
        :returns: None
        :rtype: None

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                project_id = 124976
                meta_json = api.project.get_meta(project_id)
                meta = sly.ProjectMeta.from_json(meta_json)
                key_id_map = KeyIdMap()

                pcd_id = 198703212
                ann_info = api.pointcloud.annotation.download(pcd_id)
                ann = sly.PointcloudAnnotation.from_json(ann_info, meta, key_id_map)
                figures = ann.figures[:5]
                api.video.figure.append_bulk(pcd_id, figures, key_id_map)
        """

        keys = []
        figures_json = []
        for figure in figures:
            keys.append(figure.key())
            figures_json.append(figure.to_json(key_id_map))

        self._append_bulk(pointcloud_id, figures_json, keys, key_id_map)

    def append_to_dataset(
        self,
        dataset_id: int,
        figures: List[PointcloudFigure],
        entity_ids: List[int],
        key_id_map: KeyIdMap,
    ) -> None:
        """
        Add pointcloud figures to Dataset annotations.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param figures: List of point cloud figures.
        :type figures: List[:class:`~supervisely.pointcloud_annotation.pointcloud_figure.PointcloudFigure`]
        :param entity_ids: List of point cloud IDs.
        :type entity_ids: List[int]
        :param key_id_map: KeyIdMap object.
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`, optional
        :rtype: None

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly
                from supervisely.geometry.cuboid_3d import Cuboid3d, Vector3d
                from supervisely.pointcloud_annotation.pointcloud_annotation import PointcloudObjectCollection
                from supervisely.pointcloud_annotation.pointcloud_figure import PointcloudFigure
                from supervisely.video_annotation.key_id_map import KeyIdMap

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()


                project_id = 17231
                dataset_id = 55875
                pointcloud_id = 19373403
                project = api.project.get_info_by_id(project_id)
                dataset = api.dataset.get_info_by_id(dataset_id)

                class_car = sly.ObjClass('car', Cuboid3d)
                classes = sly.ObjClassCollection([class_car])
                project_meta = sly.ProjectMeta(classes)
                updated_meta = api.project.update_meta(project.id, project_meta.to_json())

                key_id_map = KeyIdMap()

                car_object = sly.PointcloudObject(class_car)
                objects_collection = PointcloudObjectCollection([car_object])

                uploaded_objects_ids = api.pointcloud_episode.object.append_to_dataset(
                    dataset.id,
                    objects_collection,
                    key_id_map,
                )

                position, rotation, dimension = Vector3d(-32.4, 33.9, -0.7), Vector3d(0., 0, 0.1), Vector3d(1.8, 3.9, 1.6)
                cuboid = Cuboid3d(position, rotation, dimension)
                figure_1 = PointcloudFigure(car_object, cuboid)

                api.pointcloud_episode.figure.append_to_dataset(
                    dataset.id,
                    [figure_1],
                    [pointcloud_id],
                    key_id_map,
                )
        """

        keys = []
        figures_json = []
        for figure, entity_id in zip(figures, entity_ids):
            keys.append(figure.key())
            figure_json = figure.to_json(key_id_map)
            figure_json[ApiField.ENTITY_ID] = entity_id
            figures_json.append(figure_json)

        return self._append_bulk(
            dataset_id, figures_json, keys, key_id_map, field_name=ApiField.DATASET_ID
        )

    def _convert_json_info(self, info: dict, skip_missing=True):
        return super()._convert_json_info(info, skip_missing)

    def download(
        self,
        dataset_id: int,
        pointcloud_ids: List[int] = None,
        skip_geometry: bool = False,
        **kwargs
    ) -> Dict[int, List[FigureInfo]]:
        """
        Method returns a dictionary with pairs of pointcloud ID and list of FigureInfo for the given dataset ID. Can be filtered by pointcloud IDs.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param pointcloud_ids: Specify the list of pointcloud IDs within the given dataset ID. If pointcloud_ids is None, the method returns all possible pairs of images with figures. Note: Consider using `sly.batched()` to ensure that no figures are lost in the response.
        :type pointcloud_ids: List[int], optional
        :param skip_geometry: Skip the download of figure geometry. May be useful for a significant api request speed increase in the large datasets.
        :type skip_geometry: bool

        :returns: A dictionary where keys are pointcloud IDs and values are lists of figures.
        :rtype: Dict[int, List[:class:`~supervisely.api.entity_annotation.figure_api.FigureInfo`]]
        """
        if kwargs.get("image_ids", False) is not False:
            pointcloud_ids = kwargs["image_ids"]  # backward compatibility
        return super().download(dataset_id, pointcloud_ids, skip_geometry)



class TaskType:
    """Types of tasks for Neural Networks."""

    OBJECT_DETECTION = "object detection"
    INSTANCE_SEGMENTATION = "instance segmentation"
    SEMANTIC_SEGMENTATION = "semantic segmentation"
    POSE_ESTIMATION = "pose estimation"
    SALIENT_OBJECT_SEGMENTATION = "salient object segmentation"
    PROMPT_BASED_OBJECT_DETECTION = "prompt-based object detection"
    INTERACTIVE_SEGMENTATION = "interactive segmentation"
    PROMPTABLE_SEGMENTATION = "promptable segmentation"
    TRACKING = "tracking"
    OBJECT_DETECTION_3D = "object detection 3d"

AVAILABLE_TASK_TYPES = [
    TaskType.OBJECT_DETECTION,
    TaskType.INSTANCE_SEGMENTATION,
    TaskType.SEMANTIC_SEGMENTATION,
    TaskType.POSE_ESTIMATION,
    TaskType.SALIENT_OBJECT_SEGMENTATION,
    TaskType.PROMPT_BASED_OBJECT_DETECTION,
    TaskType.INTERACTIVE_SEGMENTATION,
    TaskType.PROMPTABLE_SEGMENTATION,
    TaskType.TRACKING,
    TaskType.OBJECT_DETECTION_3D,
]


from collections import defaultdict
from typing import List

import numpy as np

from supervisely.nn.benchmark.base_visualizer import BaseVisMetrics
from supervisely.nn.benchmark.visualization.widgets import ChartWidget
from supervisely.nn.task_type import TaskType


class OutcomeCounts(BaseVisMetrics):
    """TP/FP/FN outcome counts and common-vs-different charts for detection comparison."""

    CHART_MAIN = "chart_outcome_counts"
    CHART_COMPARISON = "chart_outcome_counts_comparison"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.imgIds_to_anns = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))
        self.coco_to_sly_ids = defaultdict(lambda: defaultdict(lambda: defaultdict(tuple)))

        self._initialize_ids_mapping()

        self.common_and_diff_tp = self._find_common_and_diff_tp()
        self.common_and_diff_fn = self._find_common_and_diff_fn()
        self.common_and_diff_fp = self._find_common_and_diff_fp()

    def _initialize_ids_mapping(self):
        for idx, r in enumerate(self.eval_results):
            l = {
                "TP": (r.mp.cocoDt.anns, r.mp.m.tp_matches, r.click_data.outcome_counts),
                "FN": (r.mp.cocoGt.anns, r.mp.m.fn_matches, r.click_data.outcome_counts),
                "FP": (r.mp.cocoDt.anns, r.mp.m.fp_matches, r.click_data.outcome_counts),
            }
            for outcome, (coco_anns, matches_data, sly_data) in l.items():
                for m, sly_m in zip(matches_data, sly_data[outcome]):
                    key = m["dt_id"] if outcome != "FN" else m["gt_id"]
                    gt_key = m["gt_id"] if outcome != "FP" else m["dt_id"]
                    ann = coco_anns[key]
                    self.imgIds_to_anns[idx][outcome][key].append(ann)
                    self.coco_to_sly_ids[idx][outcome][gt_key] = sly_m

    @property
    def chart_widget_main(self) -> ChartWidget:
        chart = ChartWidget(name=self.CHART_MAIN, figure=self.get_main_figure())
        chart.set_click_data(
            gallery_id=self.explore_modal_table.id,
            click_data=self.get_main_click_data(),
            chart_click_extra="'getKey': (payload) => `${payload.points[0].y}${'_'}${payload.points[0].data.name}`,",
        )
        return chart

    @property
    def chart_widget_comparison(self) -> ChartWidget:
        chart = ChartWidget(
            name=self.CHART_COMPARISON,
            figure=self.get_comparison_figure(),
        )
        chart.set_click_data(
            gallery_id=self.explore_modal_table.id,
            click_data=self.get_comparison_click_data(),
            chart_click_extra="'getKey': (payload) => `${payload.points[0].y}${'_'}${payload.points[0].data.name}`,",
        )
        return chart

    def _update_figure_layout(self, fig):
        fig.update_layout(
            barmode="stack",
            width=600,
            height=300,
        )
        fig.update_xaxes(title_text="Count (objects)")
        fig.update_yaxes(tickangle=-90)

        fig.update_layout(
            dragmode=False,
            modebar=dict(
                remove=[
                    "zoom2d",
                    "pan2d",
                    "select2d",
                    "lasso2d",
                    "zoomIn2d",
                    "zoomOut2d",
                    "autoScale2d",
                    "resetScale2d",
                ]
            ),
        )
        return fig

    def get_main_figure(self):  #  -> Optional[go.Figure]
        import plotly.graph_objects as go  # pylint: disable=import-error

        fig = go.Figure()
        tp_counts = [eval_result.mp.TP_count for eval_result in self.eval_results][::-1]
        fn_counts = [eval_result.mp.FN_count for eval_result in self.eval_results][::-1]
        fp_counts = [eval_result.mp.FP_count for eval_result in self.eval_results][::-1]
        model_names = [f"[{i}] {e.short_name}" for i, e in enumerate(self.eval_results, 1)][::-1]
        counts = [tp_counts, fn_counts, fp_counts]
        names = ["TP", "FN", "FP"]
        colors = ["#8ACAA1", "#dd3f3f", "#F7ADAA"]

        for metric, values, color in zip(names, counts, colors):
            fig.add_trace(
                go.Bar(
                    x=values,
                    y=model_names,
                    name=metric,
                    orientation="h",
                    marker=dict(color=color),
                    hovertemplate=f"{metric}: %{{x}} objects<extra></extra>",
                )
            )

        fig = self._update_figure_layout(fig)
        return fig

    def get_comparison_figure(self):  #  -> Optional[go.Figure]
        import plotly.graph_objects as go  # pylint: disable=import-error

        fig = go.Figure()

        colors = ["#8ACAA1", "#dd3f3f", "#F7ADAA"]
        model_names = [f"[{i}] {e.short_name}" for i, e in enumerate(self.eval_results, 1)][::-1]
        model_names.append("Common")

        diff_tps, common_tps = self.common_and_diff_tp
        diff_fns, common_fns = self.common_and_diff_fn
        diff_fps, common_fps = self.common_and_diff_fp
        tps_cnt = [len(x) for x in diff_tps[::-1]] + [len(common_tps)]
        fns_cnt = [len(x) for x in diff_fns[::-1]] + [len(common_fns)]
        fps_cnt = [len(x) for x in diff_fps[::-1]] + [len(common_fps)]

        for metric, values, color in zip(["TP", "FN", "FP"], [tps_cnt, fns_cnt, fps_cnt], colors):
            fig.add_trace(
                go.Bar(
                    x=values,
                    y=model_names,
                    name=metric,
                    orientation="h",
                    marker=dict(color=color),
                    hovertemplate=f"{metric}: %{{x}} objects<extra></extra>",
                )
            )

        fig = self._update_figure_layout(fig)
        return fig

    def _find_common_and_diff_fn(self) -> List[int]:
        ids = [
            dict([(x["gt_obj_id"], x["gt_img_id"]) for x in r.click_data.outcome_counts["FN"]])
            for r in self.eval_results
        ]
        same = set.intersection(*[set(x.keys()) for x in ids])
        diffs = [set(x.keys()) - same for x in ids]

        same = {i: ids[0][i] for i in same}
        diffs = [{i: s[i] for i in d} for s, d in zip(ids, diffs)]

        return diffs, same

    def _get_coco_key_name(self):
        task_type_to_key_name = {
            TaskType.OBJECT_DETECTION: "bbox",
            TaskType.INSTANCE_SEGMENTATION: "segmentation",
            TaskType.SEMANTIC_SEGMENTATION: "segmentation",
        }
        key_name = task_type_to_key_name.get(self.eval_results[0].cv_task)
        if key_name is None:
            raise NotImplementedError("Not implemented for this task type")
        return key_name

    def _find_common_and_diff_fp(self) -> List[int]:
        from pycocotools import mask as maskUtils  # pylint: disable=import-error

        iouThr = 0.75
        key_name = self._get_coco_key_name()

        imgIds_to_anns = [self.imgIds_to_anns[idx]["FP"] for idx in range(len(self.eval_results))]
        sly_ids_list = [
            {x["dt_obj_id"]: x["dt_img_id"] for x in r.click_data.outcome_counts["FP"]}
            for r in self.eval_results
        ]

        same_fp_matches = []
        for img_id in imgIds_to_anns[0]:
            anns_list = [imgIds[img_id] for imgIds in imgIds_to_anns]
            geoms_list = [[x[key_name] for x in anns] for anns in anns_list]

            if any(len(geoms) == 0 for geoms in geoms_list):
                continue

            ious_list = [
                maskUtils.iou(geoms_list[0], geoms, [0] * len(geoms)) for geoms in geoms_list[1:]
            ]
            if any(len(ious) == 0 for ious in ious_list):
                continue

            indxs_list = [np.nonzero(ious > iouThr) for ious in ious_list]
            if any(len(indxs[0]) == 0 for indxs in indxs_list):
                continue

            indxs_list = [list(zip(*indxs)) for indxs in indxs_list]
            indxs_list = [
                sorted(indxs, key=lambda x: ious[x[0], x[1]], reverse=True)
                for indxs, ious in zip(indxs_list, ious_list)
            ]

            id_sets = [set(idxs[0]) for idxs in indxs_list]
            common_ids = set.intersection(*id_sets)
            if not common_ids:
                continue

            for i, j in indxs_list[0]:
                if i in common_ids:
                    same_fp_matches.append((anns_list[0][i], [anns[j] for anns in anns_list[1:]]))
                    common_ids.remove(i)

        # Find different FP matches for each model
        same_fp_ids = set(x[0]["id"] for x in same_fp_matches)
        diff_fp_matches = [
            set([x["dt_id"] for x in eval_result.mp.m.fp_matches]) - same_fp_ids
            for eval_result in self.eval_results
        ]

        diff_fp_matches_dicts = []
        for idx, diff_fp in enumerate(diff_fp_matches):
            diff_fp_dict = {}
            for x in diff_fp:
                obj_id = self.coco_to_sly_ids[idx]["FP"][x]["dt_obj_id"]
                img_id = sly_ids_list[idx][obj_id]
                diff_fp_dict[obj_id] = img_id
            diff_fp_matches_dicts.append(diff_fp_dict)

        same_fp_matches_dict = {}
        for x in same_fp_matches:
            obj_id = self.coco_to_sly_ids[0]["FP"][x[0]["id"]]["dt_obj_id"]
            img_id = sly_ids_list[0][obj_id]
            same_fp_matches_dict[obj_id] = img_id

        return diff_fp_matches_dicts, same_fp_matches_dict

    def _find_common_and_diff_tp(self) -> tuple:

        ids = [
            dict([(x["gt_obj_id"], x) for x in r.click_data.outcome_counts["TP"]])
            for r in self.eval_results
        ]

        same = set.intersection(*[set(x.keys()) for x in ids])
        diffs = [set(x.keys()) - same for x in ids]

        same = {s["dt_obj_id"]: s["dt_img_id"] for s in [ids[0][i] for i in same]}
        diffs = [{s[i]["dt_obj_id"]: s[i]["dt_img_id"] for i in d} for s, d in zip(ids, diffs)]

        return diffs, same

    def get_main_click_data(self):
        res = {}

        res["layoutTemplate"] = [None, None, None]
        res["clickData"] = {}
        for i, eval_result in enumerate(self.eval_results, 1):
            model_name = f"[{i}] {eval_result.name}"
            for outcome, matches_data in eval_result.click_data.outcome_counts.items():
                key = f"{model_name}_{outcome}"
                outcome_dict = res["clickData"].setdefault(key, {})
                outcome_dict["imagesIds"] = []

                img_ids = set()
                obj_ids = set()
                for x in matches_data:
                    img_ids.add(x["dt_img_id"] if outcome != "FN" else x["gt_img_id"])
                    obj_ids.add(x["dt_obj_id"] if outcome != "FN" else x["gt_obj_id"])

                title = f"{model_name}. {outcome}: {len(obj_ids)} object{'s' if len(obj_ids) > 1 else ''}"
                outcome_dict["title"] = title
                outcome_dict["imagesIds"] = list(img_ids)
                thr = eval_result.mp.conf_threshold
                if outcome == "FN":
                    outcome_dict["filters"] = [
                        {"type": "specific_objects", "tagId": None, "value": list(obj_ids)},
                    ]
                else:
                    outcome_dict["filters"] = [
                        {"type": "tag", "tagId": "outcome", "value": outcome},
                        {"type": "tag", "tagId": "confidence", "value": [thr, 1]},
                    ]

        return res

    def get_comparison_click_data(self):
        res = {}

        res["layoutTemplate"] = [None, None, None]

        res["clickData"] = {}

        outcomes_ids = {
            "TP": self.common_and_diff_tp,
            "FN": self.common_and_diff_fn,
            "FP": self.common_and_diff_fp,
        }

        def _update_outcome_dict(title, outcome, outcome_dict, ids):
            img_ids = set()
            obj_ids = set()
            for obj_id, img_id in ids.items():
                img_ids.add(img_id)
                obj_ids.add(obj_id)

            title = f"{title}. {outcome}: {len(obj_ids)} object{'s' if len(obj_ids) > 1 else ''}"
            outcome_dict["title"] = title
            outcome_dict["imagesIds"] = list(img_ids)
            filters = outcome_dict.setdefault("filters", [])
            filters.append({"type": "specific_objects", "tagId": None, "value": list(obj_ids)})
            if outcome != "FN":
                filters.append({"type": "tag", "tagId": "confidence", "value": [0, 1]})
                filters.append({"type": "tag", "tagId": "outcome", "value": outcome})

        for outcome, (diff_ids, common_ids) in outcomes_ids.items():
            key = f"Common_{outcome}"
            outcome_dict = res["clickData"].setdefault(key, {})

            _update_outcome_dict("Common", outcome, outcome_dict, common_ids)

            for i, diff_ids in enumerate(diff_ids, 1):
                name = f"[{i}] {self.eval_results[i - 1].name}"
                key = f"{name}_{outcome}"
                outcome_dict = res["clickData"].setdefault(key, {})

                _update_outcome_dict(name, outcome, outcome_dict, diff_ids)

        return res
